from srb.core.action import (
    ActionGroup,
    JointPositionActionCfg,
    JointPositionActionGroup,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, LeggedRobot, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class Cassie(LeggedRobot):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/cassie",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("legged")
            .joinpath("cassie.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.9),
            joint_pos={
                "hip_abduction_left": 0.1,
                "hip_rotation_left": 0.0,
                "hip_flexion_left": 1.0,
                "thigh_joint_left": -1.8,
                "ankle_joint_left": 1.57,
                "toe_joint_left": -1.57,
                "hip_abduction_right": -0.1,
                "hip_rotation_right": 0.0,
                "hip_flexion_right": 1.0,
                "thigh_joint_right": -1.8,
                "ankle_joint_right": 1.57,
                "toe_joint_right": -1.57,
            },
            joint_vel={".*": 0.0},
        ),
        actuators={
            "legs": ImplicitActuatorCfg(
                joint_names_expr=["hip_.*", "thigh_.*", "ankle_.*"],
                effort_limit=200.0,
                velocity_limit=10.0,
                stiffness={
                    "hip_abduction.*": 100.0,
                    "hip_rotation.*": 100.0,
                    "hip_flexion.*": 200.0,
                    "thigh_joint.*": 200.0,
                    "ankle_joint.*": 200.0,
                },
                damping={
                    "hip_abduction.*": 3.0,
                    "hip_rotation.*": 3.0,
                    "hip_flexion.*": 6.0,
                    "thigh_joint.*": 6.0,
                    "ankle_joint.*": 6.0,
                },
            ),
            "toes": ImplicitActuatorCfg(
                joint_names_expr=["toe_.*"],
                effort_limit=20.0,
                velocity_limit=10.0,
                stiffness={
                    "toe_joint.*": 20.0,
                },
                damping={
                    "toe_joint.*": 1.0,
                },
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="pelvis")
    frame_payload_mount: Frame = Frame(
        prim_relpath="pelvis",
        offset=Transform(
            pos=(-0.125, 0.0, 0.1),
            rot=rpy_to_quat(0.0, -30.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="pelvis",
        offset=Transform(
            pos=(0.1, 0.0, 0.125),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="pelvis/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = "(left|right)_toe"



from enum import Enum, auto

from typing_extensions import Self


class SceneryType(str, Enum):
    EXTRAVEHICULAR = auto()
    INTRAVEHICULAR = auto()
    TERRAIN = auto()
    SUBTERRANE = auto()

    def __str__(self) -> str:
        return self.name.lower()

    @classmethod
    def from_str(cls, string: str) -> Self | None:
        return next(
            (variant for variant in cls if string.upper() == variant.name), None
        )



from srb.core.asset import Object, RigidObjectCfg
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


### ANCHOR: example (docs)
class SampleTube(Object):
    ## Model - Rigid object affected by physics
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/sample",
        ## Spawner loads a static USD file
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("sample_tube.usdc").as_posix()
            ),
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=1500.0),
        ),
    )
    ### ANCHOR_END: example (docs)



import simforge_foundry

from srb.core.asset import Object, RigidObjectCfg
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    PreviewSurfaceCfg,
    RigidBodyPropertiesCfg,
    SimforgeAssetCfg,
    UsdFileCfg,
)
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


class Peg(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/peg",
        spawn=SimforgeAssetCfg(
            assets=[simforge_foundry.PegGeo()],
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(mesh_approximation="sdf"),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=1000.0),
        ),
    )


class Hole(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/hole",
        spawn=SimforgeAssetCfg(
            assets=[simforge_foundry.HoleGeo()],
            collision_props=CollisionPropertiesCfg(),
            rigid_props=RigidBodyPropertiesCfg(kinematic_enabled=True),
        ),
    )


class ProfilePeg(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/peg",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("peg_in_hole")
                .joinpath("profile_simple.usdc")
                .as_posix()
            ),
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="boundingCube"
            ),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=1000.0),
            visual_material=PreviewSurfaceCfg(diffuse_color=(0.6, 0.6, 0.6)),
        ),
    )


class ShortProfilePeg(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/peg",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("peg_in_hole")
                .joinpath("profile_short_simple.usdc")
                .as_posix()
            ),
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="boundingCube"
            ),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=1000.0),
            visual_material=PreviewSurfaceCfg(diffuse_color=(0.6, 0.6, 0.6)),
        ),
    )


class ProfileHole(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/hole",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("peg_in_hole")
                .joinpath("hole_simple.usdc")
                .as_posix()
            ),
            collision_props=CollisionPropertiesCfg(),
            rigid_props=RigidBodyPropertiesCfg(kinematic_enabled=True),
            visual_material=PreviewSurfaceCfg(diffuse_color=(0.6, 0.6, 0.6)),
        ),
    )



from math import pi

from srb.core.action import (
    ActionGroup,
    BinaryJointVelocityActionCfg,
    JointVelocityBinaryActionGroup,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import (
    ActiveTool,
    ArticulationCfg,
    Frame,
    RigidObjectCfg,
    Tool,
    Transform,
)
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT, SRB_ASSETS_DIR_SRB_ROBOT


class ManualScrewdriverM3(Tool):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/screwdriver",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("screwdriver_hex_m3.usdz").as_posix()
            ),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(mesh_approximation="sdf"),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=1500.0),
        ),
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="screwdriver")
    frame_tool_centre_point: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.075)))


class ManualScrewdriverM5(Tool):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/screwdriver",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("screwdriver_hex_m5.usdz").as_posix()
            ),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(mesh_approximation="sdf"),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=1500.0),
        ),
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="screwdriver")
    frame_tool_centre_point: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.075)))


class ElectricScrewdriverM3(ActiveTool):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/screwdriver",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("screwdriver")
            .joinpath("electric_screwdriver_hex_m3.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
        ),
        actuators={
            "driver": ImplicitActuatorCfg(
                joint_names_expr=["driver_joint", "counterweight_joint"],
                velocity_limit=4.0 * pi,
                effort_limit=100.0,
                stiffness=0.0,
                damping=5000.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointVelocityBinaryActionGroup(
        BinaryJointVelocityActionCfg(
            asset_name="robot",
            joint_names=["driver_joint", "counterweight_joint"],
            open_command_expr={"driver_joint": 0.0, "counterweight_joint": 0.0},
            close_command_expr={
                "driver_joint": 4.0 * pi,
                "counterweight_joint": -4.0 * pi,
            },
        )
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="body")
    frame_tool_centre_point: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.075)))


class ElectricScrewdriverM5(ActiveTool):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/screwdriver",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("screwdriver")
            .joinpath("electric_screwdriver_hex_m5.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
        ),
        actuators={
            "driver": ImplicitActuatorCfg(
                joint_names_expr=["driver_joint", "counterweight_joint"],
                velocity_limit=4.0 * pi,
                effort_limit=100.0,
                stiffness=0.0,
                damping=5000.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointVelocityBinaryActionGroup(
        BinaryJointVelocityActionCfg(
            asset_name="robot",
            joint_names=["driver_joint", "counterweight_joint"],
            open_command_expr={"driver_joint": 0.0, "counterweight_joint": 0.0},
            close_command_expr={
                "driver_joint": 4.0 * pi,
                "counterweight_joint": -4.0 * pi,
            },
        )
    )

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="body")
    frame_tool_centre_point: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.075)))



from typing import TYPE_CHECKING

import elements
from embodied import Driver

if TYPE_CHECKING:
    from srb._typing import AnyEnvCfg


class DriverParallelEnv(Driver):
    def __init__(self, env: "AnyEnvCfg", num_envs: int, **kwargs):
        self.kwargs = kwargs
        self.length = num_envs
        self.env = env
        self.act_space = self.env.act_space  # type: ignore
        self.callbacks = []
        self.acts = None
        self.carry = None
        self.reset()

    def close(self):
        self.env.close()

    def _step(self, policy, step, episode):
        acts = self.acts
        obs = self.env.step(acts)  # type: ignore
        logs = {k: v for k, v in obs.items() if k.startswith("log/")}  # type: ignore
        obs = {k: v for k, v in obs.items() if not k.startswith("log/")}  # type: ignore
        self.carry, acts, outs = policy(self.carry, obs, **self.kwargs)
        if obs["is_last"].any():
            mask = ~obs["is_last"]
            acts = {k: self._mask(v, mask) for k, v in acts.items()}
        self.acts = {**acts, "reset": obs["is_last"].copy()}
        trans = {**obs, **acts, **outs, **logs}
        for i in range(self.length):
            trn = elements.tree.map(lambda x: x[i], trans)
            [fn(trn, i, **self.kwargs) for fn in self.callbacks]
        step += len(obs["is_first"])
        episode += obs["is_last"].sum()
        return step, episode



from typing import Callable, Tuple, Union

from srb.core.visuals import rtx_post
from srb.utils.cfg import configclass


def set_visuals(cfg: "VisualsCfg"):
    rtx_post.auto_exposure(**cfg.auto_exposure.to_dict())  # type: ignore
    rtx_post.chromatic_aberration(**cfg.chromatic_aberration.to_dict())  # type: ignore
    rtx_post.depth_of_field(**cfg.depth_of_field.to_dict())  # type: ignore
    rtx_post.fog(**cfg.fog.to_dict())  # type: ignore
    rtx_post.lens_flare(**cfg.lens_flare.to_dict())  # type: ignore
    rtx_post.motion_blur(**cfg.motion_blur.to_dict())  # type: ignore
    rtx_post.reshade(**cfg.reshade.to_dict())  # type: ignore
    rtx_post.tv_noise(**cfg.tv_noise.to_dict())  # type: ignore


@configclass
class FogCfg:
    enable: bool = False
    color: Tuple[float, float, float] = (0.75, 0.75, 0.75)
    intensity: float = 1.0
    z_up: bool = True
    start_height: float = 1.0
    height_density: float = 1.0
    height_falloff: float = 1.0
    distance_range: Tuple[float, float] = (0.0, 1024.0)
    fog_distance_density: float = 1.0


@configclass
class AutoExposureCfg:
    enable: bool = False
    filter_type: int = 0  # 0 (Median) | 1 (Average)
    tau: float = 3.5
    white_scale: float = 10.0
    use_exposure_clamping: bool = True
    min_ev: float = 50.0
    max_ev: float = 100000.0


@configclass
class ChromaticAberrationCfg:
    enable: bool = False
    strength: Tuple[float, float, float] = (-0.055, -0.075, 0.015)
    mode: Union[Tuple[int, int, int], int] = (0, 0, 0)  # 0 (Radial) | 1 (Barrel)
    lanczos: bool = False


@configclass
class DepthOfFieldCfg:
    enable: bool = False
    subject_distance: float = 2.0
    focal_length: float = 35.0
    f_number: float = 5.0
    anisotropy: float = 0.0


@configclass
class MotionBlurCfg:
    enable: bool = False
    diameter_fraction: float = 0.02
    exposure_fraction: float = 1.0
    num_samples: int = 8


@configclass
class LensFlareCfg:
    enable: bool = False
    scale: float = 1.0
    cutoff_point: Union[Tuple[float, float, float], float] = (2.0, 2.0, 2.0)
    cutoff_fuzziness: float = 0.5
    alpha_exposure_scale: float = 1.0
    energy_constraining_blend: bool = False
    blades: int = 5
    aperture_rotation: float = 5.0
    sensor_diagonal: float = 60.0
    sensor_aspect_ratio: float = 1.5
    f_number: float = 5.0
    focal_length: float = 35.0
    noise_strength: float = 0.0
    dust_strength: float = 0.0
    scratch_strength: float = 0.0
    spectral_blur_samples: int = 0
    spectral_blur_intensity: float = 10.0
    spectral_blur_wavelength_range: Tuple[float, float, float] = (380.0, 550.0, 770.0)


@configclass
class TVNoiseCfg:
    enable_scanlines: bool = False
    scanline_spread: float = 1.0
    enable_scroll_bug: bool = False
    enable_vignetting: bool = False
    vignetting_size: float = 107.0
    vignetting_strength: float = 0.7
    enable_vignetting_flickering: bool = False
    enable_ghost_flickering: bool = False
    enable_wave_distortion: bool = False
    enable_vertical_lines: bool = False
    enable_random_splotches: bool = False
    enable_film_grain: bool = False
    grain_amount: float = 0.05
    color_amount: float = 0.6
    lum_amount: float = 1.0
    grain_size: float = 1.6


@configclass
class ReshadeCfg:
    enable: bool = False
    preset_file_path: str = ""
    effect_search_dir_path: str = ""
    texture_search_dir_path: str = "/root/isaac-sim/kit/reshade"


@configclass
class VisualsCfg:
    func: Callable = set_visuals

    fog: FogCfg = FogCfg()
    auto_exposure: AutoExposureCfg = AutoExposureCfg()
    chromatic_aberration: ChromaticAberrationCfg = ChromaticAberrationCfg()
    depth_of_field: DepthOfFieldCfg = DepthOfFieldCfg()
    motion_blur: MotionBlurCfg = MotionBlurCfg()
    lens_flare: LensFlareCfg = LensFlareCfg()
    tv_noise: TVNoiseCfg = TVNoiseCfg()
    reshade: ReshadeCfg = ReshadeCfg()



import torch

from srb.core.action.action_group import ActionGroup
from srb.core.action.term import BodyAccelerationActionCfg
from srb.utils.cfg import configclass


@configclass
class BodyAccelerationActionGroup(ActionGroup):
    body_vel: BodyAccelerationActionCfg = BodyAccelerationActionCfg(asset_name="robot")

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        return twist


@configclass
class BodyAccelerationRelativeActionGroup(ActionGroup):
    body_vel: BodyAccelerationActionCfg = BodyAccelerationActionCfg(
        asset_name="robot", relative=True
    )

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        return twist



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset.object.object import Object, ObjectRegistry
from srb.core.asset.object.object_type import ObjectType


class Payload(Object, object_entrypoint=ObjectType.PAYLOAD):
    @classmethod
    def object_registry(cls) -> Sequence[Type[Payload]]:
        return ObjectRegistry.registry.get(ObjectType.PAYLOAD, [])  # type: ignore



import bisect
import math
import types
from dataclasses import MISSING
from os import environ
from typing import Dict, Literal, get_type_hints

import torch
from simforge import BakeType

from srb import assets
from srb.core.action import (
    ActionGroup,
    ActionTermCfg,
    DifferentialInverseKinematicsActionCfg,
    OperationalSpaceControllerActionCfg,
)
from srb.core.action.term import DummyActionCfg
from srb.core.asset import (
    ActiveTool,
    ArticulationCfg,
    AssetBaseCfg,
    AssetVariant,
    CombinedMobileManipulator,
    Manipulator,
    MobileRobot,
    RigidObjectCfg,
    RigidObjectCollectionCfg,
    Robot,
    Scenery,
    Tool,
)
from srb.core.domain import Domain
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.marker import VisualizationMarkersCfg
from srb.core.mdp import follow_xform_orientation_linear_trajectory  # noqa F401
from srb.core.mdp import reset_scene_to_default  # noqa F401
from srb.core.mdp import reset_xform_orientation_uniform
from srb.core.sensor import SensorBaseCfg
from srb.core.sim import (
    DistantLightCfg,
    DomeLightCfg,
    ParticlesSpawnerCfg,
    PhysxCfg,
    PyramidParticlesSpawnerCfg,
    RenderCfg,
    RigidBodyMaterialCfg,
    SimforgeAssetCfg,
    SimulationCfg,
)
from srb.core.sim.robot_setup import RobotAssemblerCfg
from srb.core.visuals import VisualsCfg
from srb.utils import logging
from srb.utils.cfg import configclass
from srb.utils.math import combine_frame_transforms_tuple, rpy_to_quat
from srb.utils.path import (
    SRB_ASSETS_DIR_SRB_SKYDOME_HIGH_RES,
    SRB_ASSETS_DIR_SRB_SKYDOME_LOW_RES,
)

from .event_cfg import BaseEventCfg
from .scene_cfg import BaseSceneCfg


@configclass
class BaseEnvCfg:
    ## Scenario
    seed: int = 0
    domain: Domain = Domain.MOON

    ## Assets
    scenery: Scenery | AssetVariant | None = AssetVariant.PROCEDURAL
    _scenery: Scenery | None = MISSING  # type: ignore
    robot: Robot | AssetVariant = AssetVariant.DATASET
    _robot: Robot = MISSING  # type: ignore
    skydome: Literal["low_res", "high_res"] | bool | None = "low_res"

    ## Assemblies (dynamic joints)
    joint_assemblies: Dict[str, RobotAssemblerCfg] = {}

    ## Scene
    scene: BaseSceneCfg = BaseSceneCfg()
    stack: bool = False
    num_envs: int | None = None
    spacing: float | None = None

    ## Events
    events: BaseEventCfg = BaseEventCfg()

    ## Time
    env_rate: float = MISSING  # type: ignore
    agent_rate: float = MISSING  # type: ignore

    ## Simulation
    sim = SimulationCfg(
        dt=MISSING,  # type: ignore
        render_interval=MISSING,  # type: ignore
        gravity=MISSING,  # type: ignore
        device="cpu",  # Note: Changed to GPU in __main__.py because initializing with CPU improves compatibility
        physx=PhysxCfg(
            min_position_iteration_count=2,
            min_velocity_iteration_count=1,
            enable_ccd=True,
            enable_stabilization=False,
            bounce_threshold_velocity=0.0,
            friction_offset_threshold=0.01,
            friction_correlation_distance=0.005,
        ),
        physics_material=RigidBodyMaterialCfg(
            static_friction=1.0,
            dynamic_friction=1.0,
            restitution=0.0,
            friction_combine_mode="multiply",
            restitution_combine_mode="multiply",
        ),
        render=RenderCfg(
            enable_translucency=True,
            enable_reflections=True,
        ),
    )
    malloc_scale: float = 1.0

    ## Visuals
    visuals: VisualsCfg = VisualsCfg()

    ## Misc
    truncate_episodes: bool = True
    extras: bool = False
    debug_vis: bool = (
        environ.get("DEBUG_VIS") or environ.get("SRB_DEBUG_VIS", "false")
    ).lower() in ("true", "1")

    ## Particles
    # Note: This option is likely to be removed in the future
    scatter_particles: bool = False
    particles_size: float = 0.025
    particles_ratio: float = 0.001

    def __post_init__(self):
        ## Scene
        if self.num_envs is not None:
            self.scene.num_envs = self.num_envs
        if self.spacing is None:
            self.spacing = self.scene.env_spacing
        self.scene.env_spacing = 0.0 if self.stack else self.spacing

        ## Assets -> Scene
        self._add_sunlight()
        self._add_skydome()
        self._add_scenery()
        self._add_robot()
        # Update seed & number of variants for procedural assets
        self._update_procedural_assets()

        ## Particles
        self._maybe_add_particles()
        self._maybe_disable_fabric_for_particles()

        ## Events
        self.events.update(self)

        ## Simulation
        self.decimation = math.floor(self.agent_rate / self.env_rate)
        self.sim.dt = self.env_rate
        self.sim.render_interval = self.decimation
        self.sim.gravity = (0.0, 0.0, -self.domain.gravity_magnitude)
        self._update_memory_allocation()

        ## Misc
        self._update_debug_vis()

    def _update_memory_allocation(self):
        _pow = math.floor(self.scene.num_envs**0.375) - 1

        self.sim.physx.gpu_max_rigid_contact_count = math.floor(
            self.malloc_scale * 2 ** min(13 + _pow, 31),
        )
        self.sim.physx.gpu_max_rigid_patch_count = math.floor(
            self.malloc_scale * 2 ** min(12 + _pow, 31),
        )
        self.sim.physx.gpu_found_lost_pairs_capacity = math.floor(
            self.malloc_scale * 2 ** min(12 + _pow, 31),
        )
        self.sim.physx.gpu_found_lost_aggregate_pairs_capacity = math.floor(
            self.malloc_scale * 2 ** min(13 + _pow, 31),
        )
        self.sim.physx.gpu_total_aggregate_pairs_capacity = math.floor(
            self.malloc_scale * 2 ** min(12 + _pow, 31),
        )
        self.sim.physx.gpu_collision_stack_size = math.floor(
            self.malloc_scale * 2 ** min(19 + _pow, 31),
        )
        self.sim.physx.gpu_heap_capacity = math.floor(
            self.malloc_scale * 2 ** min(19 + _pow, 31),
        )
        self.sim.physx.gpu_temp_buffer_capacity = math.floor(
            self.malloc_scale * 2 ** min(10 + _pow, 31),
        )
        self.sim.physx.gpu_max_soft_body_contacts = math.floor(
            self.malloc_scale * 2 ** min(10 + _pow, 31),
        )
        self.sim.physx.gpu_max_particle_contacts = math.floor(
            self.malloc_scale * 2 ** min(10 + _pow, 31),
        )

        self.sim.physx.gpu_max_num_partitions = 1 << bisect.bisect_left(
            (3, 15, 127, 511, 1023), self.scene.num_envs
        )

    def _add_sunlight(self, *, prim_path: str = "/World/sunlight", **kwargs):
        if self.domain.light_intensity <= 0.0:
            return
        self.scene.sunlight = AssetBaseCfg(
            prim_path=prim_path,
            spawn=DistantLightCfg(
                intensity=self.domain.light_intensity,
                angle=self.domain.light_angular_diameter,
                color_temperature=self.domain.light_color_temperature,
                enable_color_temperature=True,
                **kwargs,
            ),
            init_state=AssetBaseCfg.InitialStateCfg(
                rot=rpy_to_quat(
                    45.0,
                    30.0,
                    0.0,
                ),
            ),
        )

    def _add_skydome(self, *, prim_path: str = "/World/skydome", **kwargs):
        if not self.skydome:
            self.scene.skydome = None
            self.events.randomize_skydome_orientation = None
            return
        elif isinstance(self.skydome, str):
            match self.skydome:
                case "low_res":
                    skydome_dir = SRB_ASSETS_DIR_SRB_SKYDOME_LOW_RES
                case "high_res":
                    skydome_dir = SRB_ASSETS_DIR_SRB_SKYDOME_HIGH_RES
                case _:
                    raise ValueError(f"Invalid skydome option: {self.skydome}")
        else:
            skydome_dir = SRB_ASSETS_DIR_SRB_SKYDOME_LOW_RES

        match self.domain:
            case Domain.EARTH:
                self.scene.skydome = AssetBaseCfg(
                    prim_path=prim_path,
                    spawn=DomeLightCfg(
                        intensity=0.25 * self.domain.light_intensity,
                        texture_file=skydome_dir.joinpath("cloudy_sky.exr").as_posix(),
                        **kwargs,
                    ),
                )
            case Domain.MOON:
                self.scene.skydome = AssetBaseCfg(
                    prim_path=prim_path,
                    spawn=DomeLightCfg(
                        intensity=0.25 * self.domain.light_intensity,
                        texture_file=skydome_dir.joinpath(
                            "stars.exr"
                            # "milky_way.exr
                        ).as_posix(),
                        **kwargs,
                    ),
                )
                self.events.randomize_skydome_orientation = EventTermCfg(
                    func=reset_xform_orientation_uniform,
                    mode="interval",
                    is_global_time=True,
                    interval_range_s=(10.0, 60.0),
                    params={
                        "asset_cfg": SceneEntityCfg("skydome"),
                        "orientation_distribution_params": {
                            "roll": (-torch.pi, torch.pi),
                            "pitch": (-torch.pi, torch.pi),
                            "yaw": (-torch.pi, torch.pi),
                        },
                    },
                )
            case Domain.MARS:
                self.scene.skydome = AssetBaseCfg(
                    prim_path=prim_path,
                    spawn=DomeLightCfg(
                        intensity=0.25 * self.domain.light_intensity,
                        texture_file=skydome_dir.joinpath("mars_sky.exr").as_posix(),
                        **kwargs,
                    ),
                )
                self.events.randomize_skydome_orientation = None
            case Domain.ORBIT:
                self.scene.skydome = AssetBaseCfg(
                    prim_path=prim_path,
                    spawn=DomeLightCfg(
                        intensity=0.25 * self.domain.light_intensity,
                        texture_file=skydome_dir.joinpath(
                            "low_earth_orbit.exr"
                            # "low_lunar_orbit.jpg"
                        ).as_posix(),
                        **kwargs,
                    ),
                )
                self.events.randomize_skydome_orientation = EventTermCfg(
                    func=reset_xform_orientation_uniform,
                    mode="interval",
                    is_global_time=True,
                    interval_range_s=(10.0, 60.0),
                    params={
                        "asset_cfg": SceneEntityCfg("skydome"),
                        "orientation_distribution_params": {
                            "roll": (-torch.pi, torch.pi),
                            "pitch": (-torch.pi, torch.pi),
                            "yaw": (-torch.pi, torch.pi),
                        },
                    },
                )
            case _:
                self.scene.skydome = None
                self.events.randomize_skydome_orientation = None

    def _add_scenery(
        self,
        *,
        prim_path: str = "{ENV_REGEX_NS}/scenery",
        prim_path_stacked: str = "/World/scenery",
        **kwargs,
    ):
        if self.scenery is None:
            self._scenery = None
            return

        ## Select scenery from registry based on the selected variant
        scenery = self.scenery
        if isinstance(scenery, AssetVariant):
            type_hints = get_type_hints(self)["scenery"]

            ## Attributes
            assert self.spacing is not None
            size = (self.spacing, self.spacing)
            scale = (
                self.spacing,
                self.spacing,
                0.1 * self.spacing,
            )
            _dyn_res = max(
                1,  # Min multiplier
                min(
                    10,  # Max multiplier
                    2
                    * round(
                        min(
                            8.0,
                            math.pow(
                                self.spacing,
                                0.4,  # Exponential scaling
                            ),
                        )
                        / 2
                    ),
                ),
            )
            texture_resolution = {
                BakeType.ALBEDO: _dyn_res * 1024,
                BakeType.EMISSION: _dyn_res * 128,
                BakeType.METALLIC: _dyn_res * 256,
                BakeType.NORMAL: _dyn_res * 1024,
                BakeType.ROUGHNESS: _dyn_res * 512,
            }
            density = 0.01 * (_dyn_res**2)
            flat_area_size = 0.8 * (_dyn_res**1.2)

            if isinstance(type_hints, types.UnionType):
                for typ in type_hints.__args__:
                    if issubclass(typ, AssetVariant):
                        continue
                    assert issubclass(typ, Scenery)
                    for registered_scenery in typ.scenery_registry():
                        if issubclass(registered_scenery, typ):
                            try:
                                _scenery = registered_scenery(
                                    size=size,
                                    scale=scale,
                                    texture_resolution=texture_resolution,
                                    density=density,
                                    flat_area_size=flat_area_size,
                                    **kwargs,
                                )
                            except Exception as e:
                                logging.warning(
                                    f'Failed to instantiate "{registered_scenery.__name__}": {e}'
                                )
                            if _scenery.is_variant(scenery):
                                if (
                                    len(_scenery.DOMAINS) == 0
                                    or self.domain in _scenery.DOMAINS
                                ):
                                    scenery = _scenery
                                    logging.info(
                                        f'Selected scenery "{scenery.__class__.__name__}" of variant "{scenery.asset_variant}" for domain "{self.domain}"'
                                    )
                                    break
                                else:
                                    logging.debug(
                                        f'Registered scenery "{registered_scenery.__name__}" is a subclass of suitable type "{typ.__name__}" and "{_scenery.asset_variant}" variant but it does not support the "{self.domain}" domain'
                                    )
                            else:
                                logging.debug(
                                    f'Registered scenery "{registered_scenery.__name__}" is a subclass of suitable type "{typ.__name__}" but its "{_scenery.asset_variant}" variant does not match the requested "{scenery}" variant'
                                )
                    else:
                        continue
                    break
            else:
                logging.error(
                    f"Unsupported type hints for scenery specified via {AssetVariant}: {type_hints} ({type(type_hints)})"
                )
        assert isinstance(scenery, Scenery), (
            f"Failed to instantiate scenery from {repr(scenery)}"
        )

        # Update prim path
        scenery.asset_cfg.prim_path = (
            prim_path_stacked
            if self.stack or isinstance(self.scenery, assets.GroundPlane)
            else prim_path
        )

        # Add to the scene
        self.scene.scenery = scenery.asset_cfg

        # Store the updated config in an internal state
        self._scenery = scenery

    def _add_robot(
        self,
        *,
        prim_path: str = "{ENV_REGEX_NS}/robot",
        prim_path_manipulator: str = "{ENV_REGEX_NS}/manipulator",
        prim_path_payload: str = "{ENV_REGEX_NS}/payload",
        prim_path_end_effector: str = "{ENV_REGEX_NS}/end_effector",
        **kwargs,
    ):
        robot_name = prim_path.rstrip("/").rsplit("/", 1)[-1]
        manipulator_name = prim_path_manipulator.rstrip("/").rsplit("/", 1)[-1]
        payload_name = prim_path_payload.rstrip("/").rsplit("/", 1)[-1]
        end_effector_name = prim_path_end_effector.rstrip("/").rsplit("/", 1)[-1]

        ## Select robot from registry based on the selected variant
        robot = self.robot
        if isinstance(robot, AssetVariant):
            type_hints = get_type_hints(self)["robot"]
            if isinstance(type_hints, types.UnionType):
                for typ in type_hints.__args__:
                    if issubclass(typ, AssetVariant):
                        continue
                    assert issubclass(typ, Robot)
                    for registered_robot in typ.robot_registry():
                        if issubclass(registered_robot, typ):
                            try:
                                _robot = registered_robot(**kwargs)
                            except Exception as e:
                                logging.warning(
                                    f'Failed to instantiate "{registered_robot.__name__}": {e}'
                                )
                            if _robot.is_variant(robot):
                                if (
                                    len(_robot.DOMAINS) == 0
                                    or self.domain in _robot.DOMAINS
                                ):
                                    robot = _robot
                                    logging.info(
                                        f'Selected robot "{robot.__class__.__name__}" of variant "{robot.asset_variant}" for domain "{self.domain}"'
                                    )
                                    break
                                else:
                                    logging.debug(
                                        f'Registered robot "{registered_robot.__name__}" is a subclass of suitable type "{typ.__name__}" and "{_robot.asset_variant}" variant but it does not support the "{self.domain}" domain'
                                    )
                            else:
                                logging.debug(
                                    f'Registered robot "{registered_robot.__name__}" is a subclass of suitable type "{typ.__name__}" but its "{_robot.asset_variant}" variant does not match the requested "{robot}" variant'
                                )
                    else:
                        continue
                    break
            else:
                logging.error(
                    f"Unsupported type hints for robot specified via {AssetVariant}: {type_hints} ({type(type_hints)})"
                )
        assert isinstance(robot, Robot), (
            f"Failed to instantiate robot from {repr(robot)}"
        )

        ## Reuse existing action group or create a new one
        # if hasattr(self, "actions") and isinstance(self.actions, ActionGroup):
        #     map_cmd_to_action_fns = [self.actions.map_cmd_to_action]
        # else:
        #     self.actions = ActionGroup()
        #     map_cmd_to_action_fns = []
        self.actions = ActionGroup()
        map_cmd_to_action_fns = []

        ## Common: Robot
        robot.asset_cfg.prim_path = prim_path
        setattr(self.scene, robot_name, robot.asset_cfg)
        # Actions
        for action_key, action_term in robot.actions.__dict__.items():
            if not isinstance(action_term, ActionTermCfg):
                continue
            # Ensure the actions terms are applied to the correct asset
            action_term.asset_name = robot_name
            # Add action terms to the action group
            setattr(self.actions, f"{robot_name}__{action_key}", action_term)
        # Add the command mapping function to the action group
        map_cmd_to_action_fns.append(robot.actions.map_cmd_to_action)

        ## Extra: Mobile robot - Payload
        if isinstance(robot, MobileRobot):
            if robot.payload is not None:
                if isinstance(
                    robot.payload.asset_cfg, (RigidObjectCfg, ArticulationCfg)
                ):
                    robot.payload.asset_cfg.prim_path = prim_path_payload
                    (
                        robot.payload.asset_cfg.init_state.pos,
                        robot.payload.asset_cfg.init_state.rot,
                    ) = combine_frame_transforms_tuple(
                        robot.asset_cfg.init_state.pos,
                        robot.asset_cfg.init_state.rot,
                        robot.frame_payload_mount.offset.pos,
                        robot.frame_payload_mount.offset.rot,
                    )
                    self.joint_assemblies[payload_name] = RobotAssemblerCfg(
                        base_path=robot.asset_cfg.prim_path,
                        attach_path=robot.payload.asset_cfg.prim_path,
                        base_mount_frame=f"/{robot.frame_payload_mount.prim_relpath}"
                        if robot.frame_payload_mount.prim_relpath
                        else "",
                        attach_mount_frame="",
                        fixed_joint_offset=robot.frame_payload_mount.offset.pos,
                        fixed_joint_orient=robot.frame_payload_mount.offset.rot,
                        mask_all_collisions=True,
                    )
                else:
                    self.joint_assemblies.pop(payload_name, None)
                    robot.payload.asset_cfg.prim_path = f"{robot.asset_cfg.prim_path}/{robot.frame_base.prim_relpath}/{payload_name}"
                    (
                        robot.payload.asset_cfg.init_state.pos,
                        robot.payload.asset_cfg.init_state.rot,
                    ) = (
                        robot.frame_payload_mount.offset.pos,
                        robot.frame_payload_mount.offset.rot,
                    )
                setattr(self.scene, payload_name, robot.payload.asset_cfg)
            else:
                setattr(self.scene, payload_name, None)
                self.joint_assemblies.pop(payload_name, None)

        ## Extra: Mobile manipulator (combined) - Manipulator
        if isinstance(robot, CombinedMobileManipulator):
            ## Manipulator
            robot.manipulator.asset_cfg.prim_path = prim_path_manipulator
            setattr(self.scene, manipulator_name, robot.manipulator.asset_cfg)
            manipulator_needs_jacobian = bool(
                next(
                    (
                        action_term
                        for action_term in robot.manipulator.actions.__dict__.values()
                        if isinstance(
                            action_term,
                            (
                                DifferentialInverseKinematicsActionCfg,
                                OperationalSpaceControllerActionCfg,
                            ),
                        )
                    ),
                    None,
                )
            )
            (
                robot.manipulator.asset_cfg.init_state.pos,
                robot.manipulator.asset_cfg.init_state.rot,
            ) = combine_frame_transforms_tuple(
                robot.asset_cfg.init_state.pos,
                robot.asset_cfg.init_state.rot,
                robot.frame_manipulator_mount.offset.pos,
                robot.frame_manipulator_mount.offset.rot,
            )
            self.joint_assemblies[manipulator_name] = RobotAssemblerCfg(
                base_path=prim_path,
                attach_path=prim_path_manipulator,
                base_mount_frame=f"/{robot.frame_base.prim_relpath}"
                if robot.frame_base.prim_relpath
                else "",
                attach_mount_frame=f"/{robot.manipulator.frame_base.prim_relpath}"
                if robot.manipulator.frame_base.prim_relpath
                else "",
                fixed_joint_offset=robot.frame_manipulator_mount.offset.pos,
                fixed_joint_orient=robot.frame_manipulator_mount.offset.rot,
                mask_all_collisions=True,
                disable_root_joints=not manipulator_needs_jacobian,
            )
            # Actions
            for (
                action_key,
                action_term,
            ) in robot.manipulator.actions.__dict__.items():
                if not isinstance(action_term, ActionTermCfg):
                    continue
                # Ensure the actions terms are applied to the correct asset
                action_term.asset_name = manipulator_name
                # Add action terms to the action group
                setattr(self.actions, f"{manipulator_name}__{action_key}", action_term)
            # Add the command mapping function to the action group
            map_cmd_to_action_fns.append(robot.manipulator.actions.map_cmd_to_action)
        else:
            self.joint_assemblies.pop(manipulator_name, None)

        ## Extra: Manipulator & Mobile manipulator (combined) - End-effector
        if isinstance(robot, (Manipulator, CombinedMobileManipulator)):
            manipulator = robot if isinstance(robot, Manipulator) else robot.manipulator
            if isinstance(manipulator.end_effector, Tool):
                if isinstance(
                    manipulator.end_effector.asset_cfg,
                    (RigidObjectCfg, ArticulationCfg),
                ):
                    manipulator.end_effector.asset_cfg.prim_path = (
                        prim_path_end_effector
                    )
                    (_fixed_joint_offset, _fixed_joint_orient) = (
                        combine_frame_transforms_tuple(
                            manipulator.frame_flange.offset.pos,
                            manipulator.frame_flange.offset.rot,
                            manipulator.end_effector.frame_mount.offset.pos,
                            manipulator.end_effector.frame_mount.offset.rot,
                        )
                    )
                    (
                        manipulator.end_effector.asset_cfg.init_state.pos,
                        manipulator.end_effector.asset_cfg.init_state.rot,
                    ) = (_fixed_joint_offset, _fixed_joint_orient)
                    self.joint_assemblies[end_effector_name] = RobotAssemblerCfg(
                        base_path=manipulator.asset_cfg.prim_path,
                        attach_path=manipulator.end_effector.asset_cfg.prim_path,
                        base_mount_frame=f"/{manipulator.frame_flange.prim_relpath}"
                        if manipulator.frame_flange.prim_relpath
                        else "",
                        attach_mount_frame=f"/{manipulator.end_effector.frame_mount.prim_relpath}"
                        if manipulator.end_effector.frame_mount.prim_relpath
                        else "",
                        fixed_joint_offset=_fixed_joint_offset,
                        fixed_joint_orient=_fixed_joint_orient,
                        mask_all_collisions=True,
                    )
                else:
                    self.joint_assemblies.pop(end_effector_name, None)
                    manipulator.end_effector.asset_cfg.prim_path = f"{manipulator.asset_cfg.prim_path}/{manipulator.frame_flange.prim_relpath}/{end_effector_name}"
                    (
                        manipulator.end_effector.asset_cfg.init_state.pos,
                        manipulator.end_effector.asset_cfg.init_state.rot,
                    ) = (
                        manipulator.frame_flange.offset.pos,
                        manipulator.frame_flange.offset.rot,
                    )
                setattr(
                    self.scene, end_effector_name, manipulator.end_effector.asset_cfg
                )
                # Offset TCP
                for action_term in manipulator.actions.__dict__.values():
                    if isinstance(
                        action_term,
                        (
                            DifferentialInverseKinematicsActionCfg,
                            OperationalSpaceControllerActionCfg,
                        ),
                    ):
                        if action_term.body_offset is None:
                            action_term.body_offset = action_term.__class__.OffsetCfg(  # type: ignore
                                pos=manipulator.end_effector.frame_tool_centre_point.offset.pos,
                                rot=manipulator.end_effector.frame_tool_centre_point.offset.rot,
                            )
                        (
                            action_term.body_offset.pos,
                            action_term.body_offset.rot,
                        ) = combine_frame_transforms_tuple(
                            action_term.body_offset.pos,
                            action_term.body_offset.rot,
                            manipulator.end_effector.frame_tool_centre_point.offset.pos,
                            manipulator.end_effector.frame_tool_centre_point.offset.rot,
                        )
                # Actions
                if isinstance(manipulator.end_effector, ActiveTool):
                    for (
                        action_key,
                        action_term,
                    ) in manipulator.end_effector.actions.__dict__.items():
                        if not isinstance(action_term, ActionTermCfg):
                            continue
                        # Ensure the actions terms are applied to the correct asset
                        action_term.asset_name = end_effector_name
                        # Add action terms to the action group
                        setattr(
                            self.actions,
                            f"{end_effector_name}__{action_key}",
                            action_term,
                        )
                    # Add the command mapping function to the action group
                    map_cmd_to_action_fns.append(
                        manipulator.end_effector.actions.map_cmd_to_action
                    )
            else:
                setattr(self.scene, end_effector_name, None)
                self.joint_assemblies.pop(end_effector_name, None)

        ## WORKAROUND: Mobile manipulator (combined) - Dummy action
        ## Note: The term is used by mobile manipulators that require Jacobian for their actions to enable/disable the root joint (kept enabled by default)
        if isinstance(robot, CombinedMobileManipulator) and manipulator_needs_jacobian:
            # Add action terms to the action group
            setattr(
                self.actions,
                f"{robot_name}_{manipulator_name}_dummy_switch",
                DummyActionCfg(),
            )
            # Add the command mapping function to the action group
            map_cmd_to_action_fns.append(
                lambda twist, event: torch.Tensor((-1.0 if event else 1.0,)).to(
                    device=twist.device
                )
            )

        ## Update command mapping function
        self.actions.map_cmd_to_action = lambda twist, event: torch.cat(
            [func(twist, event) for func in map_cmd_to_action_fns]
        )

        # Store the updated config in an internal state
        self._robot = robot

    def _update_procedural_assets(self):
        def _update_simforge_asset(asset_cfg: AssetBaseCfg):
            if not isinstance(asset_cfg.spawn, SimforgeAssetCfg):
                return
            if asset_cfg.spawn.seed == 0:
                asset_cfg.spawn.seed = self.seed
            if asset_cfg.spawn.num_assets == 1 and (
                asset_cfg.prim_path.startswith("{ENV_REGEX_NS}")
                or asset_cfg.prim_path.startswith("/World/envs/env_.*")
            ):
                asset_cfg.spawn.num_assets = self.scene.num_envs

        for asset_cfg in self.scene.__dict__.values():
            if isinstance(asset_cfg, AssetBaseCfg):
                _update_simforge_asset(asset_cfg)
            elif isinstance(asset_cfg, RigidObjectCollectionCfg):
                if not isinstance(asset_cfg.rigid_objects, Dict):
                    continue
                for rigid_object_cfg in asset_cfg.rigid_objects.values():
                    _update_simforge_asset(rigid_object_cfg)

    def _maybe_add_particles(self):
        assert self.spacing is not None
        if self.scatter_particles and self.spacing > 0.0:
            self.scene.particles = AssetBaseCfg(  # type: ignore
                prim_path="{ENV_REGEX_NS}/particles",
                spawn=PyramidParticlesSpawnerCfg(
                    ratio=self.particles_ratio,
                    particle_size=self.particles_size,
                    dim_x=round(self.spacing / self.particles_size),
                    dim_y=round(self.spacing / self.particles_size),
                    dim_z=round(0.5 * self.spacing / self.particles_size),
                    velocity=((-0.5, 0.5), (-0.5, 0.5), (-0.5, 0.0)),
                    fluid=False,
                    density=1500.0,
                    friction=0.85,
                    cohesion=0.65,
                ),
                init_state=AssetBaseCfg.InitialStateCfg(pos=(0.0, 0.0, 0.5)),
            )

    def _maybe_disable_fabric_for_particles(self):
        for asset_cfg in self.scene.__dict__.values():
            if isinstance(asset_cfg, AssetBaseCfg) and isinstance(
                asset_cfg.spawn, ParticlesSpawnerCfg
            ):
                self.sim.use_fabric = False
                return
            elif isinstance(asset_cfg, RigidObjectCollectionCfg):
                if not isinstance(asset_cfg.rigid_objects, Dict):
                    continue
                for rigid_object_cfg in asset_cfg.rigid_objects.values():
                    if isinstance(rigid_object_cfg, AssetBaseCfg) and isinstance(
                        rigid_object_cfg.spawn, ParticlesSpawnerCfg
                    ):
                        self.sim.use_fabric = False
                        return

    def _update_debug_vis(self):
        for asset_cfg in self.scene.__dict__.values():
            if isinstance(asset_cfg, SensorBaseCfg):
                asset_cfg.debug_vis = self.debug_vis

        for action_template in self.actions.__dict__.values():
            if isinstance(action_template, ActionTermCfg):
                action_template.debug_vis = self.debug_vis

        for attr in self.__dict__.values():
            if isinstance(attr, VisualizationMarkersCfg):
                for marker in attr.markers.values():
                    marker.visible = self.debug_vis



from typing import Any, Dict

from pydantic import model_validator

from srb.core.asset.robot.manipulation import Manipulator
from srb.core.asset.robot.mobile.mobile_robot import MobileRobot
from srb.core.asset.robot.mobile_manipulation.mobile_manipulator import (
    MobileManipulator,
)


class CombinedMobileManipulator(
    MobileManipulator, MobileRobot, mobile_manipulator_metaclass=True, extra="allow"
):
    ## Model
    mobile_base: MobileRobot | None = None
    manipulator: Manipulator

    @model_validator(mode="before")
    @classmethod
    def override_mobile_base(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        if isinstance(data.get("mobile_base"), MobileRobot):
            for key, value in data["mobile_base"]._model_values.items():
                data[key] = value
            # data["mobile_base"] = None
        return data

    def model_post_init(self, __context):
        if isinstance(self.mobile_base, MobileRobot):
            for key, value in self.mobile_base._model_values.items():
                setattr(self, key, value)
            # self.mobile_base = None
        super().model_post_init(__context)



from dataclasses import MISSING

import torch

from srb.core.action.action_group import ActionGroup
from srb.core.action.term import WheeledDriveActionCfg
from srb.utils.cfg import configclass


@configclass
class WheeledDriveActionGroup(ActionGroup):
    cmd_vel: WheeledDriveActionCfg = MISSING  # type: ignore

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        return twist[:2]



from srb.core.action import (
    ActionGroup,
    DifferentialIKControllerCfg,
    DifferentialInverseKinematicsActionCfg,
    InverseKinematicsActionGroup,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, SerialManipulator, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.nucleus import ISAACLAB_NUCLEUS_DIR


class UnitreeZ1(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/unitree_z1",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/Unitree/Z1/z1.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "joint1": 0.0,
                "joint2": deg_to_rad(90.0),
                "joint3": deg_to_rad(-45.0),
                "joint4": deg_to_rad(45.0),
                "joint5": deg_to_rad(0.0),
                "joint6": deg_to_rad(0.0),
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=100.0,
                effort_limit=87.0,
                stiffness=800.0,
                damping=40.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=["joint.*"],
            base_name="link00",
            body_name="link06",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.05,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="link00")
    frame_flange: Frame = Frame(
        prim_relpath="link06",
        offset=Transform(pos=(0.05, 0.0, 0.0), rot=rpy_to_quat(0.0, 90.0, 0.0)),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="link00/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="link06/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )



import os
from collections import OrderedDict
from pprint import pprint
from typing import TYPE_CHECKING, Any, Dict, Tuple

import yaml
from gymnasium import spaces
from rl_zoo3.exp_manager import ExperimentManager as __ExperimentManager
from stable_baselines3.common.preprocessing import (
    is_image_space,
    is_image_space_channels_first,
)
from stable_baselines3.common.vec_env import (
    VecEnv,
    VecFrameStack,
    VecTransposeImage,
    is_vecenv_wrapped,
)

if TYPE_CHECKING:
    from srb._typing import AnyEnv


class ExperimentManager(__ExperimentManager):
    def __init__(self, *args, env: "AnyEnv", tensorboard_log: str = "", **kwargs):
        super().__init__(*args, tensorboard_log=tensorboard_log, **kwargs)
        self._env = env

        self.tensorboard_log = tensorboard_log
        self.log_path = self.log_folder
        self.save_path = os.path.join(self.log_folder, "ckpt")
        self.params_path = self.log_folder

    def read_hyperparameters(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        hyperparams = self.custom_hyperparams
        saved_hyperparams = OrderedDict(
            [(key, hyperparams[key]) for key in sorted(hyperparams.keys())]  # type: ignore
        )
        print(
            "Default hyperparameters for environment (ones being tuned will be overridden):"
        )
        pprint(saved_hyperparams)
        return hyperparams, saved_hyperparams  # type: ignore

    def _save_config(self, saved_hyperparams: Dict[str, Any]) -> None:
        # Save hyperparams
        with open(os.path.join(self.params_path, "config.yml"), "w") as f:
            yaml.dump(saved_hyperparams, f)

        print(f"Log path: {self.save_path}")

    def create_envs(
        self, n_envs: int, eval_env: bool = False, no_log: bool = False
    ) -> VecEnv:
        # Special case for GoalEnvs: log success rate too
        if (
            # env.is_goal_env  # TODO[low]: Add goal env handling for SB3/SBX
            False and len(self.monitor_kwargs) == 0
        ):
            self.monitor_kwargs = dict(info_keywords=("is_success",))

        # Wrap the env into a VecNormalize wrapper if needed
        # and load saved statistics when present
        env = self._maybe_normalize(
            self._env,  # type: ignore
            eval_env,
        )

        # Optional Frame-stacking
        if self.frame_stack is not None:
            n_stack = self.frame_stack
            env = VecFrameStack(env, n_stack)
            if self.verbose > 0:
                print(f"Stacking {n_stack} frames")

        if not is_vecenv_wrapped(env, VecTransposeImage):
            wrap_with_vectranspose = False
            if isinstance(env.observation_space, spaces.Dict):
                # If even one of the keys is an image-space in need of transpose, apply transpose
                # If the image spaces are not consistent (for instance, one is channel first,
                # the other channel last); VecTransposeImage will throw an error
                for space in env.observation_space.spaces.values():
                    wrap_with_vectranspose = wrap_with_vectranspose or (
                        is_image_space(space)
                        and not is_image_space_channels_first(space)  # type: ignore[arg-type]
                    )
            else:
                wrap_with_vectranspose = is_image_space(
                    env.observation_space
                ) and not is_image_space_channels_first(
                    env.observation_space  # type: ignore[arg-type]
                )

            if wrap_with_vectranspose:
                if self.verbose >= 1:
                    print("Wrapping the env in a VecTransposeImage.")
                env = VecTransposeImage(env)

        return env



from srb.core.marker import VisualizationMarkersCfg
from srb.core.sim import ConeCfg

CONE_CFG = VisualizationMarkersCfg(
    markers={
        "cone": ConeCfg(
            radius=0.1,
            height=1.0,
        ),
    }
)



from dataclasses import MISSING
from typing import List, Sequence

import torch

from srb import assets
from srb._typing import StepReturn
from srb.core.asset import AssetVariant, Humanoid, LeggedRobot
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.mdp import push_by_setting_velocity  # noqa: F401
from srb.core.mdp import reset_joints_by_scale
from srb.core.sensor import ContactSensor, ContactSensorCfg
from srb.utils.cfg import configclass
from srb.utils.math import matrix_from_quat, rotmat_to_rot6d, scale_transform

from .task import EventCfg, SceneCfg, Task, TaskCfg

##############
### Config ###
##############


@configclass
class LocomotionSceneCfg(SceneCfg):
    contacts_robot: ContactSensorCfg = ContactSensorCfg(
        prim_path=MISSING,  # type: ignore
        update_period=0.0,
        history_length=3,
        track_air_time=True,
    )


@configclass
class LocomotionEventCfg(EventCfg):
    randomize_robot_joints: EventTermCfg = EventTermCfg(
        func=reset_joints_by_scale,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("robot"),
            "position_range": (0.5, 1.5),
            "velocity_range": (0.0, 0.0),
        },
    )
    # push_robot: EventTermCfg = EventTermCfg(
    #     func=push_by_setting_velocity,
    #     mode="interval",
    #     interval_range_s=(10.0, 15.0),
    #     params={
    #         "asset_cfg": SceneEntityCfg("robot"),
    #         "velocity_range": {
    #             "x": (-0.5, 0.5),
    #             "y": (-0.5, 0.5),
    #         },
    #     },
    # )


@configclass
class LocomotionTaskCfg(TaskCfg):
    ## Assets
    robot: LeggedRobot | Humanoid | AssetVariant = assets.Spot()
    _robot: LeggedRobot = MISSING  # type: ignore

    ## Scene
    scene: LocomotionSceneCfg = LocomotionSceneCfg()

    ## Events
    events: LocomotionEventCfg = LocomotionEventCfg()

    ## Time
    env_rate: float = 1.0 / 125.0

    def __post_init__(self):
        super().__post_init__()

        # Sensor: Robot contacts
        self.scene.contacts_robot.prim_path = f"{self.scene.robot.prim_path}/.*"


############
### Task ###
############


class LocomotionTask(Task):
    cfg: LocomotionTaskCfg

    def __init__(self, cfg: LocomotionTaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._contacts_robot: ContactSensor = self.scene["contacts_robot"]

        ## Cache metrics
        self._feet_indices, _ = self._robot.find_bodies(
            self.cfg._robot.regex_feet_links
        )
        _all_body_indices, _ = self._robot.find_bodies(".*")
        self._undesired_contact_body_indices = [
            idx for idx in _all_body_indices if idx not in self._feet_indices
        ]

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

    def extract_step_return(self) -> StepReturn:
        ## Visualize target
        self._target_marker.visualize(self._goal)

        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Root
            tf_pos_robot=self._robot.data.root_pos_w,
            tf_quat_robot=self._robot.data.root_quat_w,
            vel_lin_robot=self._robot.data.root_lin_vel_b,
            vel_ang_robot=self._robot.data.root_ang_vel_b,
            projected_gravity_robot=self._robot.data.projected_gravity_b,
            # Transforms (world frame)
            tf_pos_target=self._goal,
            # Joints
            joint_pos_robot=self._robot.data.joint_pos,
            joint_pos_limits_robot=(
                self._robot.data.soft_joint_pos_limits
                if torch.all(torch.isfinite(self._robot.data.soft_joint_pos_limits))
                else None
            ),
            joint_acc_robot=self._robot.data.joint_acc,
            joint_applied_torque_robot=self._robot.data.applied_torque,
            # Contacts
            contact_forces_robot=self._contacts_robot.data.net_forces_w,  # type: ignore
            contact_robot=self._contacts_robot.compute_first_contact(self.step_dt),
            contact_last_air_time=self._contacts_robot.data.last_air_time,  # type: ignore
            # IMU
            imu_lin_acc=self._imu_robot.data.lin_acc_b,
            imu_ang_vel=self._imu_robot.data.ang_vel_b,
            ## Robot descriptors
            robot_feet_indices=self._feet_indices,
            robot_undesired_contact_body_indices=self._undesired_contact_body_indices,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Root
    tf_pos_robot: torch.Tensor,
    tf_quat_robot: torch.Tensor,
    vel_lin_robot: torch.Tensor,
    vel_ang_robot: torch.Tensor,
    projected_gravity_robot: torch.Tensor,
    # Transforms (world frame)
    tf_pos_target: torch.Tensor,
    # Joints
    joint_pos_robot: torch.Tensor,
    joint_pos_limits_robot: torch.Tensor | None,
    joint_acc_robot: torch.Tensor,
    joint_applied_torque_robot: torch.Tensor,
    # Contacts
    contact_forces_robot: torch.Tensor,
    contact_robot: torch.Tensor,
    contact_last_air_time: torch.Tensor,
    # IMU
    imu_lin_acc: torch.Tensor,
    imu_ang_vel: torch.Tensor,
    ## Robot descriptors
    robot_feet_indices: List[int],
    robot_undesired_contact_body_indices: List[int],
) -> StepReturn:
    num_envs = episode_length.size(0)
    # dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Root
    tf_rotmat_robot = matrix_from_quat(tf_quat_robot)
    tf_rot6d_robot = rotmat_to_rot6d(tf_rotmat_robot)

    ## Transforms (world frame)
    # Robot -> Target
    tf_pos_robot_to_target = tf_pos_robot[:, :2] - tf_pos_target[:, :2]
    dist_robot_to_target = torch.norm(tf_pos_robot_to_target, dim=-1)

    ## Joints
    joint_pos_robot_normalized = (
        scale_transform(
            joint_pos_robot,
            joint_pos_limits_robot[:, :, 0],
            joint_pos_limits_robot[:, :, 1],
        )
        if joint_pos_limits_robot is not None
        else joint_pos_robot
    )

    ## Contacts
    contact_forces_mean_robot = contact_forces_robot.mean(dim=1)

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Joint torque
    WEIGHT_JOINT_TORQUE = -0.000025
    MAX_JOINT_TORQUE_PENALTY = -4.0
    penalty_joint_torque = torch.clamp_min(
        WEIGHT_JOINT_TORQUE
        * torch.sum(torch.square(joint_applied_torque_robot), dim=1),
        min=MAX_JOINT_TORQUE_PENALTY,
    )

    # Penalty: Joint acceleration
    WEIGHT_JOINT_ACCELERATION = -0.00000025
    MAX_JOINT_ACCELERATION_PENALTY = -2.0
    penalty_joint_acceleration = torch.clamp_min(
        WEIGHT_JOINT_ACCELERATION * torch.sum(torch.square(joint_acc_robot), dim=1),
        min=MAX_JOINT_ACCELERATION_PENALTY,
    )

    # Penalty: Undesired robot contacts
    WEIGHT_UNDESIRED_ROBOT_CONTACTS = -2.0
    THRESHOLD_UNDESIRED_ROBOT_CONTACTS = 1.0
    penalty_undesired_robot_contacts = WEIGHT_UNDESIRED_ROBOT_CONTACTS * (
        torch.max(
            torch.norm(
                contact_forces_robot[:, robot_undesired_contact_body_indices, :],
                dim=-1,
            ),
            dim=1,
        )[0]
        > THRESHOLD_UNDESIRED_ROBOT_CONTACTS
    )

    # Reward: Distance | Robot <--> Target (precision)
    WEIGHT_DISTANCE_ROBOT_TO_TARGET_PRECISION = 64.0
    TANH_STD_DISTANCE_ROBOT_TO_TARGET_PRECISION = 0.05
    reward_distance_robot_to_target_precision = (
        WEIGHT_DISTANCE_ROBOT_TO_TARGET_PRECISION
        * (
            1.0
            - torch.tanh(
                dist_robot_to_target / TANH_STD_DISTANCE_ROBOT_TO_TARGET_PRECISION
            )
        )
    )

    # Reward: Feet air time
    WEIGHT_FEET_AIR_TIME = 0.5
    THRESHOLD_FEET_AIR_TIME = 0.1
    reward_feet_air_time = (
        WEIGHT_FEET_AIR_TIME
        * (dist_robot_to_target > THRESHOLD_FEET_AIR_TIME)
        * torch.sum(
            (contact_last_air_time[:, robot_feet_indices] - 0.5)
            * contact_robot[:, robot_feet_indices],
            dim=1,
        )
    )

    # Penalty: Minimize non-command motion (linear)
    WEIGHT_UNDESIRED_LIN_VEL_Z = -0.5
    penalty_undesired_lin_vel_z = WEIGHT_UNDESIRED_LIN_VEL_Z * torch.square(
        vel_lin_robot[:, 2]
    )

    # Penalty: Minimize non-command motion (angular)
    WEIGHT_UNDESIRED_ANG_VEL_XY = -0.1
    penalty_undesired_ang_vel_xy = WEIGHT_UNDESIRED_ANG_VEL_XY * torch.sum(
        torch.square(vel_ang_robot[:, :2]), dim=-1
    )

    # Penalty: Minimize rotation with the gravity direction
    WEIGHT_GRAVITY_ROTATION_ALIGNMENT = -2.0
    penalty_gravity_rotation_alignment = WEIGHT_GRAVITY_ROTATION_ALIGNMENT * (
        torch.sum(torch.square(projected_gravity_robot[:, :2]), dim=1)
        + torch.square(projected_gravity_robot[:, 2] + 1.0)
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "contact_forces_mean_robot": contact_forces_mean_robot,
                "tf_rot6d_robot": tf_rot6d_robot,
                "vel_lin_robot": vel_lin_robot,
                "vel_ang_robot": vel_ang_robot,
                "projected_gravity_robot": projected_gravity_robot,
                "tf_pos_robot_to_target": tf_pos_robot_to_target,
            },
            "state_dyn": {
                "contact_forces_robot": contact_forces_robot,
            },
            "proprio": {
                "imu_lin_acc": imu_lin_acc,
                "imu_ang_vel": imu_ang_vel,
            },
            "proprio_dyn": {
                "joint_pos_robot_normalized": joint_pos_robot_normalized,
                "joint_acc_robot": joint_acc_robot,
                "joint_applied_torque_robot": joint_applied_torque_robot,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_joint_torque": penalty_joint_torque,
            "penalty_joint_acceleration": penalty_joint_acceleration,
            "penalty_undesired_robot_contacts": penalty_undesired_robot_contacts,
            "reward_distance_robot_to_target_precision": reward_distance_robot_to_target_precision,
            "reward_feet_air_time": reward_feet_air_time,
            "penalty_undesired_lin_vel_z": penalty_undesired_lin_vel_z,
            "penalty_undesired_ang_vel_xy": penalty_undesired_ang_vel_xy,
            "penalty_gravity_rotation_alignment": penalty_gravity_rotation_alignment,
        },
        termination,
        truncation,
    )



from functools import cache
from importlib.util import find_spec
from os import environ
from pathlib import Path
from typing import Literal


@cache
def enable_ros2_bridge(
    distro: str = "humble",
    rmw_implementation: Literal["rmw_fastrtps_cpp", "rmw_cyclonedds_cpp"]
    | str = "rmw_fastrtps_cpp",
) -> bool:
    from srb.utils import logging

    ## Skip if ROS 2 is already sourced
    if find_spec("rclpy"):
        logging.debug(
            'ROS 2 is already sourced in the current environment, so "ros2_bridge" does not need to be enabled'
        )
        return True

    ## Update environment
    _append_ld_library_path(_get_ros2_bridge_lib_path(distro))
    if not environ.get("RMW_IMPLEMENTATION"):
        environ["RMW_IMPLEMENTATION"] = rmw_implementation

    ## Enable extension
    from isaacsim.core.utils.extensions import enable_extension

    assert enable_extension("isaacsim.ros2.bridge")

    ## Check if ROS 2 is now available
    if not find_spec("rclpy"):
        logging.error(
            'ROS 2 Python client library "rclpy" is still not available after trying to enable the "ros2_bridge" extension'
        )
        return False

    return True


@cache
def _get_ros2_bridge_lib_path(distro: str) -> Path:
    from srb.utils.isaacsim import get_isaacsim_path

    ros2_lib_path = (
        Path(get_isaacsim_path())
        .joinpath("exts")
        .joinpath("isaacsim.ros2.bridge")
        .joinpath(distro)
        .joinpath("lib")
    )
    assert ros2_lib_path.exists()
    return ros2_lib_path


@cache
def _append_ld_library_path(lib_path: Path):
    ld_library_path = environ.get("LD_LIBRARY_PATH", "")
    if ld_library_path:
        if lib_path.as_posix() in ld_library_path:
            return
        ld_library_path = f"{ld_library_path}:".replace("::", ":")
    environ["LD_LIBRARY_PATH"] = ld_library_path + lib_path.as_posix()



from .cfg import (  # noqa: F401
    GridParticlesSpawnerCfg,
    ParticlesSpawnerCfg,
    PyramidParticlesSpawnerCfg,
)
from .utils import (  # noqa: F401
    get_particles_pos_w,
    get_particles_vel_w,
    set_particles_pos_w,
    set_particles_vel_w,
)



from functools import partial as bind
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Literal

import elements
import embodied
import numpy
import portal
from dreamerv3 import agent as dreamer_agent
from dreamerv3 import main as dreamer_main
from isaacsim.simulation_app import SimulationApp
from ruamel import yaml

from srb.integrations.dreamer.eval import eval_only
from srb.integrations.dreamer.train import train
from srb.integrations.dreamer.wrapper import EmbodiedEnvWrapper
from srb.utils import logging
from srb.utils.cfg import stamp_dir

if TYPE_CHECKING:
    from srb._typing import AnyEnv, AnyEnvCfg

ALGO_NAME = "dreamer"
UPSTREAM_CONFIG_PATH = Path(dreamer_agent.__file__).parent.joinpath("configs.yaml")


def run(
    workflow: Literal["train", "eval"],
    env: "AnyEnv",
    sim_app: SimulationApp,
    env_id: str,
    env_cfg: "AnyEnvCfg",
    agent_cfg: dict,
    logdir: Path,
    model: Path,
    continue_training: bool | None = None,
    **kwargs,
):
    save_replay = agent_cfg.get("replay", {}).pop("save", False)

    # Determine checkpoint path
    if model:
        from_checkpoint = model
    elif workflow == "eval" or continue_training:
        from_checkpoint = logdir.joinpath("ckpt").joinpath(
            logdir.joinpath("ckpt").joinpath("latest").read_text().strip()
        )
    else:
        from_checkpoint = ""
    if from_checkpoint:
        logging.info(f"Loading model from {from_checkpoint}")

    # Special handling for eval workflow
    if workflow == "eval":
        logdir = stamp_dir(logdir.joinpath("eval"))

    # Setup logdir
    logdir = elements.Path(logdir)  # type: ignore
    logdir.mkdir()
    print("Agent logdir:", logdir)

    # Load the config
    configs: Dict[str, Any] = yaml.YAML(typ="safe").load(
        UPSTREAM_CONFIG_PATH.read_text()
    )
    config = elements.Config(configs["defaults"])
    config = config.update(
        {
            **agent_cfg,
            "seed": env_cfg.seed,
            "task": env_id.replace("/", "_"),
            "logdir": logdir,
            "run.from_checkpoint": from_checkpoint,
            "run.envs": env_cfg.scene.num_envs,  # type: ignore
            "run.eval_envs": env_cfg.scene.num_envs,  # type: ignore
        }
    )

    # Save the config
    config.save(logdir / "config.yaml")

    # Wrap the environment
    env = EmbodiedEnvWrapper(env)  # type: ignore
    for name, space in env.act_space.items():  # type: ignore
        if not space.discrete:
            env = embodied.wrappers.NormalizeAction(env, name)  # type: ignore
    env = embodied.wrappers.UnifyDtypes(env)  # type: ignore
    for name, space in env.act_space.items():  # type: ignore
        if not space.discrete:
            env = embodied.wrappers.ClipAction(env, name)  # type: ignore

    # Setup the workflow
    def init():
        elements.timer.global_timer.enabled = config.logger.timer  # type: ignore

    portal.setup(
        errfile=config.errfile and logdir / "error",
        clientkw=dict(logging_color="cyan"),
        serverkw=dict(logging_color="cyan"),
        initfns=[init],
        ipv6=config.ipv6,
    )

    args = elements.Config(
        **config.run,
        replica=config.replica,
        replicas=config.replicas,
        logdir=config.logdir,
        batch_size=config.batch_size,
        batch_length=config.batch_length,
        report_length=config.report_length,
        consec_train=config.consec_train,
        consec_report=config.consec_report,
        replay_context=config.replay_context,
    )

    def make_env():
        return env

    def make_agent(config):
        obs_space = {k: v for k, v in env.obs_space.items() if not k.startswith("log/")}  # type: ignore
        act_space = {k: v for k, v in env.act_space.items() if k != "reset"}  # type: ignore
        if config.random_agent:
            return embodied.RandomAgent(obs_space, act_space)
        cpdir = elements.Path(config.logdir)
        cpdir = cpdir.parent if config.replicas > 1 else cpdir
        return dreamer_agent.Agent(
            obs_space,
            act_space,
            elements.Config(
                **config.agent,
                logdir=config.logdir,
                seed=config.seed,
                jax=config.jax,
                batch_size=config.batch_size,
                batch_length=config.batch_length,
                replay_context=config.replay_context,
                report_length=config.report_length,
                replica=config.replica,
                replicas=config.replicas,
            ),
        )

    # Run the workflow
    match workflow:
        case "train":
            train(
                bind(make_agent, config),
                bind(make_replay, config, "replay" if save_replay else None),
                make_env,
                bind(dreamer_main.make_stream, config),
                bind(dreamer_main.make_logger, config),
                args,
                sim_app=sim_app,
            )
        case "eval":
            eval_only(
                bind(make_agent, config),
                make_env,
                bind(dreamer_main.make_logger, config),
                args,
                sim_app=sim_app,
            )


def make_replay(config, folder: str | Path | None, mode: str = "train"):
    batlen = config.batch_length if mode == "train" else config.report_length
    consec = config.consec_train if mode == "train" else config.consec_report
    capacity = config.replay.size if mode == "train" else config.replay.size / 10
    length = consec * batlen + config.replay_context
    assert config.batch_size * length <= capacity

    if folder:
        directory = elements.Path(config.logdir) / folder
        if config.replicas > 1:
            directory /= f"{config.replica:05}"
    else:
        directory = None
    replay_kwargs = {
        "length": length,
        "capacity": int(capacity),
        "online": config.replay.online,
        "chunksize": config.replay.chunksize,
        "directory": directory,
    }

    if mode == "train":
        assert (
            config.replay.fracs.uniform
            + config.replay.fracs.priority
            + config.replay.fracs.recency
            == 1.0
        ), "Replay fractions must sum to 1."

        if config.replay.fracs.priority > 0.0:
            assert config.jax.compute_dtype in ("bfloat16", "float32"), (
                "Gradient scaling for low-precision training can produce invalid loss "
                "outputs that are incompatible with prioritized replay."
            )
        if config.replay.fracs.recency > 0.0:
            recency = 1.0 / numpy.arange(1, capacity + 1) ** config.replay.recexp

        if config.replay.fracs.uniform == 1.0:
            replay_kwargs["selector"] = embodied.replay.selectors.Uniform(
                seed=config.seed
            )
        elif config.replay.fracs.priority == 1.0:
            replay_kwargs["selector"] = embodied.replay.selectors.Prioritized(
                seed=config.seed, **config.replay.prio
            )
        elif config.replay.fracs.recency == 1.0:
            replay_kwargs["selector"] = embodied.replay.selectors.Recency(
                recency, seed=config.seed
            )
        else:
            from srb.integrations.dreamer.selector import Mixture

            replay_kwargs["selector"] = Mixture(
                selectors={
                    "uniform": embodied.replay.selectors.Uniform(seed=config.seed),
                    "priority": embodied.replay.selectors.Prioritized(
                        seed=config.seed, **config.replay.prio
                    ),
                    "recency": embodied.replay.selectors.Recency(
                        recency, seed=config.seed
                    ),
                },
                fractions=config.replay.fracs,
                seed=config.seed,
            )

    return embodied.replay.Replay(**replay_kwargs)



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset.robot.manipulation.manipulator import (
    Manipulator,
    ManipulatorRegistry,
)
from srb.core.asset.robot.manipulation.manipulator_type import ManipulatorType


class SerialManipulator(Manipulator, manipulator_entrypoint=ManipulatorType.SERIAL):
    @classmethod
    def manipulator_registry(cls) -> Sequence[Type[SerialManipulator]]:
        return ManipulatorRegistry.registry.get(ManipulatorType.SERIAL, [])  # type: ignore



from typing import TYPE_CHECKING, Tuple

from srb import assets
from srb.core.asset import AssetVariant, RigidObjectCfg

if TYPE_CHECKING:
    from .task import TaskCfg


def select_debris(
    env_cfg: "TaskCfg",
    *,
    prim_path: str = "{ENV_REGEX_NS}/debris",
    scale: Tuple[float, float, float] = (0.05, 0.05, 0.05),
    init_state: RigidObjectCfg.InitialStateCfg = RigidObjectCfg.InitialStateCfg(),
    **kwargs,
) -> RigidObjectCfg:
    match env_cfg.debris:
        case AssetVariant.PRIMITIVE:
            debris_cfg = assets.RandomShape(scale=scale).asset_cfg

        case AssetVariant.DATASET:
            debris_cfg = assets.ProfilePeg().asset_cfg

        case AssetVariant.PROCEDURAL:
            debris_cfg = assets.Cubesat(scale=scale).asset_cfg

    debris_cfg.prim_path = prim_path
    debris_cfg.init_state = init_state
    debris_cfg.spawn.replace(**kwargs)  # type: ignore

    return debris_cfg



from srb.core.action import ActionGroup, JointPositionActionGroup
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ActiveTool, ArticulationCfg, Frame, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.nucleus import ISAAC_NUCLEUS_DIR


class AllegroHand(ActiveTool):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/allegro_hand",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/AllegroHand/allegro_hand_instanceable.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                retain_accelerations=False,
                enable_gyroscopic_forces=False,
                angular_damping=0.01,
                max_linear_velocity=1000.0,
                max_angular_velocity=3665,
                max_depenetration_velocity=1000.0,
                max_contact_impulse=1e32,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.5),
            rot=(0.257551, 0.283045, 0.683330, -0.621782),
            joint_pos={"^(?!thumb_joint_0).*": 0.0, "thumb_joint_0": 0.28},
        ),
        actuators={
            "fingers": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                effort_limit=0.5,
                velocity_limit=100.0,
                stiffness=3.0,
                damping=0.1,
                friction=0.01,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup()

    ## Frames
    frame_mount: Frame = Frame(prim_relpath="allegro_mount")
    frame_tool_centre_point: Frame = Frame(
        prim_relpath="allegro_mount", offset=Transform(pos=(0.0, 0.0, 0.15))
    )



from .env import (  # noqa: F401
    GroundManipulationEnv,
    GroundManipulationEnvCfg,
    GroundManipulationEventCfg,
    GroundManipulationSceneCfg,
)
from .visual_ext import GroundManipulationEnvVisualExtCfg  # noqa: F401



from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Dict, Iterable, List, Sequence, Tuple, Type

from srb.core.asset import ArticulationCfg, RigidObjectCfg
from srb.core.asset.robot.mobile_manipulation.mobile_manipulator_type import (
    MobileManipulatorType,
)
from srb.core.asset.robot.robot import Robot, RobotRegistry
from srb.core.asset.robot.robot_type import RobotType


class MobileManipulator(Robot, robot_entrypoint=RobotType.MOBILE_MANIPULATOR):
    ## Model
    asset_cfg: RigidObjectCfg | ArticulationCfg

    def __init_subclass__(
        cls,
        mobile_manipulator_entrypoint: MobileManipulatorType | None = None,
        mobile_manipulator_metaclass: bool = False,
        robot_metaclass: bool = False,
        **kwargs,
    ):
        super().__init_subclass__(
            robot_metaclass=(
                robot_metaclass
                or mobile_manipulator_entrypoint is not None
                or mobile_manipulator_metaclass
            ),
            **kwargs,
        )
        if mobile_manipulator_entrypoint is not None:
            assert isinstance(mobile_manipulator_entrypoint, MobileManipulatorType), (
                f"Class '{cls.__name__}' is marked as a mobile manipulator entrypoint, but '{mobile_manipulator_entrypoint}' is not a valid {MobileManipulatorType}"
            )
            assert (
                mobile_manipulator_entrypoint
                not in MobileManipulatorRegistry.base_types.keys()
            ), (
                f"Class '{cls.__name__}' is marked as '{mobile_manipulator_entrypoint}' mobile manipulator entrypoint, but it was already marked by '{MobileManipulatorRegistry.base_types[mobile_manipulator_entrypoint].__name__}'"
            )
            MobileManipulatorRegistry.base_types[mobile_manipulator_entrypoint] = cls
        elif mobile_manipulator_metaclass:
            MobileManipulatorRegistry.meta_types.append(cls)
        elif not robot_metaclass:
            for (
                mobile_manipulator_type,
                base,
            ) in MobileManipulatorRegistry.base_types.items():
                if issubclass(cls, base):
                    if (
                        mobile_manipulator_type
                        not in MobileManipulatorRegistry.registry.keys()
                    ):
                        MobileManipulatorRegistry.registry[mobile_manipulator_type] = []
                    else:
                        assert cls.name() not in (
                            mobile_manipulator.name()
                            for mobile_manipulator in MobileManipulatorRegistry.registry[
                                mobile_manipulator_type
                            ]
                        ), (
                            f"Cannot register multiple mobile manipulators with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(mobile_manipulator for mobile_manipulator in MobileManipulatorRegistry.registry[mobile_manipulator_type] if cls.name() == mobile_manipulator.name()).__module__}:{cls.__name__}'"
                        )
                    MobileManipulatorRegistry.registry[mobile_manipulator_type].append(
                        cls
                    )
                    break

    @cached_property
    def mobile_manipulator_type(self) -> MobileManipulatorType:
        for (
            mobile_manipulator_type,
            base,
        ) in MobileManipulatorRegistry.base_types.items():
            if isinstance(self, base):
                return mobile_manipulator_type
        raise ValueError(
            f"Class '{self.__class__.__name__}' has unknown mobile manipulator type"
        )

    @classmethod
    def mobile_manipulator_registry(cls) -> Sequence[Type[MobileManipulator]]:
        return list(MobileManipulatorRegistry.values_inner())

    @classmethod
    def robot_registry(cls) -> Sequence[Type[MobileManipulator]]:
        return RobotRegistry.registry.get(RobotType.MOBILE_MANIPULATOR, [])  # type: ignore


class MobileManipulatorRegistry:
    registry: ClassVar[Dict[MobileManipulatorType, List[Type[MobileManipulator]]]] = {}
    base_types: ClassVar[Dict[MobileManipulatorType, Type[MobileManipulator]]] = {}
    meta_types: ClassVar[List[Type[MobileManipulator]]] = []

    @classmethod
    def keys(cls) -> Iterable[MobileManipulatorType]:
        return cls.registry.keys()

    @classmethod
    def items(
        cls,
    ) -> Iterable[Tuple[MobileManipulatorType, Sequence[Type[MobileManipulator]]]]:
        return cls.registry.items()

    @classmethod
    def values(cls) -> Iterable[Iterable[Type[MobileManipulator]]]:
        return cls.registry.values()

    @classmethod
    def values_inner(cls) -> Iterable[Type[MobileManipulator]]:
        return (
            mobile_manipulator
            for mobile_manipulators in cls.registry.values()
            for mobile_manipulator in mobile_manipulators
        )

    @classmethod
    def n_robots(cls) -> int:
        return sum(
            len(mobile_manipulators) for mobile_manipulators in cls.registry.values()
        )

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {
            mobile_manipulator.__module__
            for mobile_manipulators in cls.registry.values()
            for mobile_manipulator in mobile_manipulators
        }

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[MobileManipulator] | None:
        for mobile_manipulator in cls.values_inner():
            if mobile_manipulator.name() == name:
                return mobile_manipulator
        return None



from typing import TYPE_CHECKING, List, Sequence

import torch
from pxr import UsdGeom, Vt

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import AssetBase


def get_particles_pos_w(
    env: "AnyEnv",
    particles: "AssetBase",
    env_ids: Sequence[int] | None = None,
) -> torch.Tensor:
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = range(_env.num_envs)
    pos_w: List[torch.Tensor] = []
    for i in env_ids:
        pos_w.append(
            torch.tensor(
                UsdGeom.Points(
                    particles._prims[i]  # type: ignore
                )
                .GetPointsAttr()
                .Get(),
                device=_env.device,
                dtype=torch.float32,
            )
        )
    return torch.stack(pos_w, dim=0)


def get_particles_vel_w(
    env: "AnyEnv",
    particles: "AssetBase",
    env_ids: Sequence[int] | None = None,
) -> torch.Tensor:
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = range(_env.num_envs)
    pos_w: List[torch.Tensor] = []
    for i in env_ids:
        pos_w.append(
            torch.tensor(
                UsdGeom.Points(
                    particles._prims[i]  # type: ignore
                )
                .GetVelocitiesAttr()
                .Get(),
                device=_env.device,
                dtype=torch.float32,
            )
        )
    return torch.stack(pos_w, dim=0)


def set_particles_pos_w(
    env: "AnyEnv",
    particles: "AssetBase",
    positions: torch.Tensor,
    env_ids: Sequence[int] | None = None,
) -> None:
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = range(_env.num_envs)
    for i in env_ids:
        UsdGeom.Points(
            particles._prims[i]  # type: ignore
        ).GetPointsAttr().Set(Vt.Vec3fArray.FromNumpy(positions[i].cpu().numpy()))


def set_particles_vel_w(
    env: "AnyEnv",
    particles: "AssetBase",
    velocities: torch.Tensor,
    env_ids: Sequence[int] | None = None,
) -> None:
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = range(_env.num_envs)
    for i in env_ids:
        UsdGeom.Points(
            particles._prims[i]  # type: ignore
        ).GetVelocitiesAttr().Set(Vt.Vec3fArray.FromNumpy(velocities[i].cpu().numpy()))



from typing import TYPE_CHECKING, Any, Dict, Tuple

from pydantic import BaseModel, NonNegativeInt
from simforge import TexResConfig

from srb import assets
from srb.core.asset import AssetVariant, RigidObjectCfg

if TYPE_CHECKING:
    from .task import TaskCfg


class PegCfg(BaseModel, arbitrary_types_allowed=True):
    asset_cfg: RigidObjectCfg
    offset_pos_ends: Tuple[
        Tuple[float, float, float],
        Tuple[float, float, float],
    ]

    ## Rotational symmetry of the peg represented as integer
    #  0: Circle (infinite symmetry)
    #  1: No symmetry (exactly one fit)
    #  n: n-fold symmetry (360/n deg between each symmetry)
    symmetry: NonNegativeInt = 1


class HoleCfg(BaseModel):
    asset_cfg: RigidObjectCfg
    offset_pos_bottom: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    offset_pos_entrance: Tuple[float, float, float]


class PegInHoleCfg(BaseModel):
    peg: PegCfg
    hole: HoleCfg


def select_peg_in_hole_assembly(
    env_cfg: "TaskCfg",
    *,
    prim_path_peg: str = "{ENV_REGEX_NS}/peg",
    prim_path_hole: str = "{ENV_REGEX_NS}/hole",
    seed: int = 0,
    init_state: RigidObjectCfg.InitialStateCfg = RigidObjectCfg.InitialStateCfg(),
    peg_kwargs: Dict[str, Any] = {},
    hole_kwargs: Dict[str, Any] = {},
    scale: Tuple[float, float, float] = (0.05, 0.05, 0.05),
    texture_resolution: TexResConfig | None = None,
    short_peg: bool = False,
    **kwargs,
) -> PegInHoleCfg:
    match env_cfg.peg:
        case AssetVariant.DATASET:
            peg_cfg = (
                assets.ShortProfilePeg() if short_peg else assets.ProfilePeg()
            ).asset_cfg
            hole_cfg = assets.ProfileHole().asset_cfg

            rot_symmetry_n = 4
            offset_pos_ends = ((0.0, 0.0, 0.0), (0.0, 0.0, 0.2))
            offset_pos_entrance = (0.0, 0.0, 0.02)

        case AssetVariant.PROCEDURAL:
            peg_cfg = assets.Peg(
                scale=scale, texture_resolution=texture_resolution
            ).asset_cfg
            hole_cfg = assets.Hole(
                scale=scale, texture_resolution=texture_resolution
            ).asset_cfg

            peg_cfg.spawn.seed = seed  # type: ignore
            hole_cfg.spawn.seed = seed  # type: ignore

            rot_symmetry_n = 1
            offset_pos_ends = ((0.0, 0.0, 0.0), (0.0, 0.0, 0.0))
            offset_pos_entrance = (0.0, 0.0, 0.0)

    peg_cfg.prim_path = prim_path_peg
    peg_kwargs.update(**kwargs)
    peg_cfg.spawn.replace(**peg_kwargs)  # type: ignore

    hole_cfg.prim_path = prim_path_hole
    hole_kwargs.update(**kwargs)
    hole_cfg.spawn.replace(**hole_kwargs)  # type: ignore

    peg_cfg.init_state = init_state
    hole_cfg.init_state = init_state

    return PegInHoleCfg(
        peg=PegCfg(
            asset_cfg=peg_cfg,
            offset_pos_ends=offset_pos_ends,
            symmetry=rot_symmetry_n,
        ),
        hole=HoleCfg(
            asset_cfg=hole_cfg,
            offset_pos_entrance=offset_pos_entrance,
        ),
    )



from typing import TYPE_CHECKING, Type

import isaaclab.utils.math as math_utils
import torch
from isaaclab.envs.mdp.actions.actions_cfg import (
    OperationalSpaceControllerActionCfg as __OperationalSpaceControllerActionCfg,
)
from isaaclab.envs.mdp.actions.task_space_actions import (
    OperationalSpaceControllerAction as __OperationalSpaceControllerAction,
)
from isaaclab.managers.action_manager import ActionTerm
from isaaclab.utils import configclass

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import Articulation


class OperationalSpaceControllerAction(__OperationalSpaceControllerAction):
    cfg: "OperationalSpaceControllerActionCfg"
    _env: "AnyEnv"
    _asset: "Articulation"

    def __init__(self, cfg: "OperationalSpaceControllerActionCfg", env: "AnyEnv"):
        super().__init__(
            cfg,
            env,  # type: ignore
        )

        if self.cfg.base_name:
            base_ids, base_names = self._asset.find_bodies(self.cfg.base_name)
            if len(base_ids) != 1:
                raise ValueError(
                    f"Expected one match for the base name: {self.cfg.base_name}. Found {len(base_ids)}: {base_names}."
                )
            self._base_idx = base_ids[0]
        else:
            self._base_idx = None

        self._asset.write_joint_stiffness_to_sim(0.0, joint_ids=self._joint_ids)
        self._asset.write_joint_damping_to_sim(0.0, joint_ids=self._joint_ids)

    @property
    def jacobian_b(self) -> torch.Tensor:
        jacobian = self.jacobian_w

        if self._base_idx:
            base_quat_w = self._asset.data.body_quat_w[:, self._base_idx]
        else:
            base_quat_w = self._asset.data.root_quat_w

        base_rot_matrix = math_utils.matrix_from_quat(math_utils.quat_inv(base_quat_w))
        jacobian[:, :3, :] = torch.bmm(base_rot_matrix, jacobian[:, :3, :])
        jacobian[:, 3:, :] = torch.bmm(base_rot_matrix, jacobian[:, 3:, :])
        return jacobian

    def _compute_ee_pose(self):
        self._ee_pose_w[:, 0:3] = self._asset.data.body_pos_w[:, self._ee_body_idx]
        self._ee_pose_w[:, 3:7] = self._asset.data.body_quat_w[:, self._ee_body_idx]

        if self._base_idx is not None:
            base_pos_w = self._asset.data.body_pos_w[:, self._base_idx]
            base_quat_w = self._asset.data.body_quat_w[:, self._base_idx]
        else:
            base_pos_w = self._asset.data.root_pos_w
            base_quat_w = self._asset.data.root_quat_w

        self._ee_pose_b_no_offset[:, 0:3], self._ee_pose_b_no_offset[:, 3:7] = (
            math_utils.subtract_frame_transforms(
                base_pos_w,
                base_quat_w,
                self._ee_pose_w[:, 0:3],
                self._ee_pose_w[:, 3:7],
            )
        )

        if self.cfg.body_offset is not None:
            self._ee_pose_b[:, 0:3], self._ee_pose_b[:, 3:7] = (
                math_utils.combine_frame_transforms(
                    self._ee_pose_b_no_offset[:, 0:3],
                    self._ee_pose_b_no_offset[:, 3:7],
                    self._offset_pos,
                    self._offset_rot,
                )
            )
        else:
            self._ee_pose_b[:] = self._ee_pose_b_no_offset

    def _compute_ee_velocity(self):
        self._ee_vel_w[:] = self._asset.data.body_vel_w[:, self._ee_body_idx, :]

        if self._base_idx is not None:
            base_vel_w = self._asset.data.body_vel_w[:, self._base_idx]
        else:
            base_vel_w = self._asset.data.root_vel_w
        relative_vel_w = self._ee_vel_w - base_vel_w

        self._ee_vel_b[:, 0:3] = math_utils.quat_rotate_inverse(
            self._asset.data.root_quat_w, relative_vel_w[:, 0:3]
        )
        self._ee_vel_b[:, 3:6] = math_utils.quat_rotate_inverse(
            self._asset.data.root_quat_w, relative_vel_w[:, 3:6]
        )

        # Account for the offset
        if self.cfg.body_offset is not None:
            # Compute offset vector in root frame
            r_offset_b = math_utils.quat_rotate(
                self._ee_pose_b_no_offset[:, 3:7], self._offset_pos
            )
            # Adjust the linear velocity to account for the offset
            self._ee_vel_b[:, :3] += torch.cross(
                self._ee_vel_b[:, 3:], r_offset_b, dim=-1
            )
            # Angular velocity is not affected by the offset

    def _compute_ee_jacobian(self):
        self._jacobian_b[:] = self.jacobian_b
        if self.cfg.body_offset is not None:
            self._jacobian_b[:, 0:3, :] += torch.bmm(
                math_utils.skew_symmetric_matrix(self._offset_pos),
                self._jacobian_b[:, 3:, :],
            )
            self._jacobian_b[:, 3:, :] = torch.bmm(
                math_utils.matrix_from_quat(self._offset_rot),
                self._jacobian_b[:, 3:, :],
            )


@configclass
class OperationalSpaceControllerActionCfg(__OperationalSpaceControllerActionCfg):
    class_type: Type[ActionTerm] = OperationalSpaceControllerAction
    base_name: str = ""



from .body import (  # noqa: F401
    BodyAccelerationActionGroup,
    BodyAccelerationRelativeActionGroup,
)
from .dummy import DummyActionGroup  # noqa: F401
from .joint import (  # noqa: F401
    JointEffortActionGroup,
    JointPositionActionGroup,
    JointPositionBinaryActionGroup,
    JointPositionBoundedActionGroup,
    JointPositionBoundedEMAActionGroup,
    JointPositionRelativeActionGroup,
    JointVelocityActionGroup,
    JointVelocityBinaryActionGroup,
)



from srb.core.asset import Object, RigidObjectCfg
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


class BeneficiationUnit(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/beneficiation_unit",
        spawn=UsdFileCfg(
            usd_path=(SRB_ASSETS_DIR_SRB_OBJECT.joinpath("PRIVATE").as_posix()),
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(kinematic_enabled=True),
            mass_props=MassPropertiesCfg(density=1000.0),
        ),
        init_state=RigidObjectCfg.InitialStateCfg(
            pos=(1.0, 3.05, 0.967),
        ),
    )



from isaacsim.robot_setup.assembler import AssembledRobot as __AssembledRobot

from .assembled_bodies import AssembledBodies


class AssembledRobot(__AssembledRobot):
    def __init__(self, assembled_robots: AssembledBodies):
        self.assembled_robots = assembled_robots

    def set_attach_path_root_joints_enabled(self, enabled: bool):
        self.assembled_robots.set_attach_path_root_joints_enabled(enabled)



from typing import Iterable, Tuple

import carb.settings


def auto_exposure(
    enable: bool,
    filter_type: int = 0,
    tau: float = 3.5,
    white_scale: float = 10.0,
    use_exposure_clamping: bool = True,
    min_ev: float = 50.0,
    max_ev: float = 100000.0,
):
    """
    filter_type: 0 (Median) | 1 (Average)
    """

    settings = carb.settings.get_settings()
    settings.set("/rtx/post/histogram/enabled", enable)
    settings.set("/rtx/post/histogram/filterType", filter_type)
    settings.set("/rtx/post/histogram/tau", tau)
    settings.set("/rtx/post/histogram/whiteScale", white_scale)
    settings.set("/rtx/post/histogram/useExposureClamping", use_exposure_clamping)
    settings.set("/rtx/post/histogram/minEV", min_ev)
    settings.set("/rtx/post/histogram/maxEV", max_ev)


def chromatic_aberration(
    enable: bool,
    strength: Tuple[float, float, float] = (-0.055, -0.075, 0.015),
    mode: Tuple[int, int, int] | int = (0, 0, 0),
    lanczos: bool = False,
):
    """
    mode: 0 (Radial) | 1 (Barrel)
    """

    if not isinstance(mode, Iterable):
        mode = (mode, mode, mode)

    settings = carb.settings.get_settings()
    settings.set("/rtx/post/chromaticAberration/enabled", enable)
    settings.set("/rtx/post/chromaticAberration/strengthR", strength[0])
    settings.set("/rtx/post/chromaticAberration/strengthG", strength[1])
    settings.set("/rtx/post/chromaticAberration/strengthB", strength[2])
    settings.set("/rtx/post/chromaticAberration/modeR", mode[0])
    settings.set("/rtx/post/chromaticAberration/modeG", mode[1])
    settings.set("/rtx/post/chromaticAberration/modeB", mode[2])
    settings.set("/rtx/post/chromaticAberration/enableLanczos", lanczos)


def depth_of_field(
    enable: bool,
    subject_distance: float = 2.0,
    focal_length: float = 35.0,
    f_number: float = 5.0,
    anisotropy: float = 0.0,
):
    settings = carb.settings.get_settings()
    settings.set("/rtx/post/dof/overrideEnabled", enable)
    settings.set("/rtx/post/dof/enabled", enable)
    settings.set("/rtx/post/dof/subjectDistance", subject_distance)
    settings.set("/rtx/post/dof/focalLength", focal_length)
    settings.set("/rtx/post/dof/fNumber", f_number)
    settings.set("/rtx/post/dof/anisotropy", anisotropy)


def fog(
    enable: bool,
    color: Tuple[float, float, float] = (0.75, 0.75, 0.75),
    intensity: float = 1.0,
    z_up: bool = True,
    start_height: float = 1.0,
    height_density: float = 1.0,
    height_falloff: float = 1.0,
    distance_range: Tuple[float, float] = (0.0, 1024.0),
    fog_distance_density: float = 1.0,
):
    settings = carb.settings.get_settings()
    settings.set("/rtx/fog/enabled", enable)
    settings.set("/rtx/fog/fogColor", color)
    settings.set("/rtx/fog/fogColorIntensity", intensity)
    settings.set("/rtx/fog/fogZup/enabled", z_up)
    settings.set("/rtx/fog/fogStartHeight", start_height)
    settings.set("/rtx/fog/fogHeightDensity", height_density)
    settings.set("/rtx/fog/fogHeightFalloff", height_falloff)
    settings.set("/rtx/fog/fogStartDist", distance_range[0])
    settings.set("/rtx/fog/fogEndDist", distance_range[1])
    settings.set("/rtx/fog/fogDistanceDensity", fog_distance_density)


def lens_flare(
    enable: bool,
    scale: float = 1.0,
    cutoff_point: Tuple[float, float, float] | float = (2.0, 2.0, 2.0),
    cutoff_fuzziness: float = 0.5,
    alpha_exposure_scale: float = 1.0,
    energy_constraining_blend: bool = False,
    blades: int = 5,
    aperture_rotation: float = 5.0,
    sensor_diagonal: float = 60.0,
    sensor_aspect_ratio: float = 1.5,
    f_number: float = 5.0,
    focal_length: float = 35.0,
    noise_strength: float = 0.0,
    dust_strength: float = 0.0,
    scratch_strength: float = 0.0,
    spectral_blur_samples: int = 0,
    spectral_blur_intensity: float = 10.0,
    spectral_blur_wavelength_range: Tuple[float, float, float] = (380.0, 550.0, 770.0),
):
    if not isinstance(cutoff_point, Iterable):
        cutoff_point = (cutoff_point, cutoff_point, cutoff_point)

    settings = carb.settings.get_settings()
    settings.set("/rtx/post/lensFlares/enabled", enable)
    settings.set("/rtx/post/lensFlares/flareScale", scale)
    settings.set("/rtx/post/lensFlares/cutoffPoint", cutoff_point)
    settings.set("/rtx/post/lensFlares/cutoffFuzziness", cutoff_fuzziness)
    settings.set("/rtx/post/lensFlares/alphaExposureScale", alpha_exposure_scale)
    settings.set(
        "/rtx/post/lensFlares/energyConstrainingBlend", energy_constraining_blend
    )
    settings.set("/rtx/post/lensFlares/physicalSettings", True)
    settings.set("/rtx/post/lensFlares/apertureShapeCircular", blades < 3)
    settings.set("/rtx/post/lensFlares/blades", blades)
    settings.set("/rtx/post/lensFlares/apertureRotation", aperture_rotation)
    settings.set("/rtx/post/lensFlares/sensorDiagonal", sensor_diagonal)
    settings.set("/rtx/post/lensFlares/sensorAspectRatio", sensor_aspect_ratio)
    settings.set("/rtx/post/lensFlares/fNumber", f_number)
    settings.set("/rtx/post/lensFlares/focalLength", focal_length)
    settings.set("/rtx/post/lensFlares/noiseStrength", noise_strength)
    settings.set("/rtx/post/lensFlares/dustStrength", dust_strength)
    settings.set("/rtx/post/lensFlares/scratchStrength", scratch_strength)
    settings.set("/rtx/post/lensFlares/spectralBlurSamples", spectral_blur_samples)
    settings.set("/rtx/post/lensFlares/spectralBlurIntensity", spectral_blur_intensity)
    settings.set(
        "/rtx/post/lensFlares/spectralBlurWavelengthRange",
        spectral_blur_wavelength_range,
    )


def motion_blur(
    enable: bool,
    diameter_fraction: float = 0.02,
    exposure_fraction: float = 1.0,
    num_samples: int = 8,
):
    settings = carb.settings.get_settings()
    settings.set("/rtx/post/motionblur/enabled", enable)
    settings.set("/rtx/post/motionblur/maxBlurDiameterFraction", diameter_fraction)
    settings.set("/rtx/post/motionblur/exposureFraction", exposure_fraction)
    settings.set("/rtx/post/motionblur/numSamples", num_samples)


def reshade(
    enable: bool,
    preset_file_path: str = "",
    effect_search_dir_path: str = "",
    texture_search_dir_path: str = "/root/isaac-sim/kit/reshade",
):
    settings = carb.settings.get_settings()
    settings.set("/rtx/reshade/enabled", enable)
    settings.set("/rtx/reshade/presetFilePath", preset_file_path)
    settings.set("/rtx/reshade/effectSearchDirPath", effect_search_dir_path)
    settings.set("/rtx/reshade/textureSearchDirPath", texture_search_dir_path)


def tv_noise(
    enable_scanlines: bool = False,
    scanline_spread: float = 1.0,
    enable_scroll_bug: bool = False,
    enable_vignetting: bool = False,
    vignetting_size: float = 107.0,
    vignetting_strength: float = 0.7,
    enable_vignetting_flickering: bool = False,
    enable_ghost_flickering: bool = False,
    enable_wave_distortion: bool = False,
    enable_vertical_lines: bool = False,
    enable_random_splotches: bool = False,
    enable_film_grain: bool = False,
    grain_amount: float = 0.05,
    color_amount: float = 0.6,
    lum_amount: float = 1.0,
    grain_size: float = 1.6,
):
    settings = carb.settings.get_settings()
    settings.set(
        "/rtx/post/tvNoise/enabled",
        enable_scanlines
        or enable_scroll_bug
        or enable_vignetting
        or enable_ghost_flickering
        or enable_wave_distortion
        or enable_vertical_lines
        or enable_random_splotches
        or enable_film_grain,
    )
    settings.set("/rtx/post/tvNoise/enableScanlines", enable_scanlines)
    settings.set("/rtx/post/tvNoise/scanlineSpread", scanline_spread)
    settings.set("/rtx/post/tvNoise/enableScrollBug", enable_scroll_bug)
    settings.set("/rtx/post/tvNoise/enableVignetting", enable_vignetting)
    settings.set("/rtx/post/tvNoise/vignettingSize", vignetting_size)
    settings.set("/rtx/post/tvNoise/vignettingStrength", vignetting_strength)
    settings.set(
        "/rtx/post/tvNoise/enableVignettingFlickering", enable_vignetting_flickering
    )
    settings.set("/rtx/post/tvNoise/enableGhostFlickering", enable_ghost_flickering)
    settings.set("/rtx/post/tvNoise/enableWaveDistortion", enable_wave_distortion)
    settings.set("/rtx/post/tvNoise/enableVerticalLines", enable_vertical_lines)
    settings.set("/rtx/post/tvNoise/enableRandomSplotches", enable_random_splotches)
    settings.set("/rtx/post/tvNoise/enableFilmGrain", enable_film_grain)
    settings.set("/rtx/post/tvNoise/grainAmount", grain_amount)
    settings.set("/rtx/post/tvNoise/colorAmount", color_amount)
    settings.set("/rtx/post/tvNoise/lumAmount", lum_amount)
    settings.set("/rtx/post/tvNoise/grainSize", grain_size)



from functools import cache
from importlib.util import find_spec
from os import environ
from pathlib import Path


@cache
def get_isaacsim_path() -> str:
    if isaac_sim_path := environ.get("ISAAC_SIM_PATH"):
        return isaac_sim_path

    standard_path = Path(environ.get("HOME", "/root")).joinpath("isaac-sim")
    if standard_path.exists():
        return standard_path.as_posix()

    return "/root/isaac-sim"


@cache
def get_isaacsim_python() -> str:
    if isaac_sim_python := environ.get("ISAAC_SIM_PYTHON"):
        return isaac_sim_python

    standard_path = (
        Path(environ.get("HOME", "/root")).joinpath("isaac-sim").joinpath("python.sh")
    )
    if standard_path.exists():
        return standard_path.as_posix()

    return "python3"


def is_isaacsim_initialized() -> bool:
    return find_spec("isaacsim.core.version") is not None


def hide_isaacsim_ui():
    import carb.settings

    settings = carb.settings.get_settings()
    settings.set("/app/window/hideUi", True)
    settings.set("/app/window/fullscreen", True)



from dataclasses import MISSING
from typing import List, Sequence

import torch

from srb._typing import StepReturn
from srb.core.env import GroundEnv, GroundEnvCfg, GroundEventCfg, GroundSceneCfg
from srb.core.manager import EventTermCfg
from srb.core.marker import VisualizationMarkers, VisualizationMarkersCfg
from srb.core.mdp import offset_pos_natural
from srb.core.sim import CylinderCfg, PreviewSurfaceCfg
from srb.utils.cfg import configclass

##############
### Config ###
##############


@configclass
class SceneCfg(GroundSceneCfg):
    pass


@configclass
class EventCfg(GroundEventCfg):
    target_pos_evolution: EventTermCfg = EventTermCfg(
        func=offset_pos_natural,
        mode="interval",
        interval_range_s=(0.2, 0.8),
        is_global_time=True,
        params={
            "env_attr_name": "_goal",
            "axes": ("x", "y"),
            "step_range": (0.05, 0.5),
            "smoothness": 0.8,
            "pos_bounds": {
                "x": MISSING,
                "y": MISSING,
            },
        },
    )


@configclass
class TaskCfg(GroundEnvCfg):
    ## Scene
    scene: SceneCfg = SceneCfg()
    stack: bool = True

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 60.0
    is_finite_horizon: bool = False

    ## Target
    target_pos_range_ratio: float = 0.9
    target_marker_cfg: VisualizationMarkersCfg = VisualizationMarkersCfg(
        prim_path="/Visuals/target",
        markers={
            "target": CylinderCfg(
                radius=0.02,
                height=50.0,
                visual_material=PreviewSurfaceCfg(emissive_color=(0.2, 0.2, 0.8)),
            )
        },
    )

    def __post_init__(self):
        super().__post_init__()

        # Event: Waypoint target
        assert self.spacing is not None
        for dim in ("x", "y"):
            self.events.target_pos_evolution.params["pos_bounds"][dim] = (  # type: ignore
                -0.5 * self.target_pos_range_ratio * self.spacing,
                0.5 * self.target_pos_range_ratio * self.spacing,
            )


############
### Task ###
############


class Task(GroundEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._target_marker: VisualizationMarkers = VisualizationMarkers(
            self.cfg.target_marker_cfg
        )

        ## Initialize buffers
        self._goal = self.scene.env_origins + torch.zeros(
            self.num_envs, 3, device=self.device
        )

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

        ## Reset goal position
        self._goal[env_ids] = self.scene.env_origins[env_ids] + torch.zeros(
            len(env_ids), 3, device=self.device
        )

    def extract_step_return(self) -> StepReturn:
        ## Visualize target
        self._target_marker.visualize(self._goal)

        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Root
            tf_pos_robot=self._robot.data.root_pos_w,
            vel_lin_robot=self._robot.data.root_lin_vel_b,
            vel_ang_robot=self._robot.data.root_ang_vel_b,
            # Transforms (world frame)
            tf_pos_target=self._goal,
            # IMU
            imu_lin_acc=self._imu_robot.data.lin_acc_b,
            imu_ang_vel=self._imu_robot.data.ang_vel_b,
            ## Robot descriptors
            forward_drive_indices=self._forward_drive_indices,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Root
    tf_pos_robot: torch.Tensor,
    vel_lin_robot: torch.Tensor,
    vel_ang_robot: torch.Tensor,
    # Transforms (world frame)
    tf_pos_target: torch.Tensor,
    # IMU
    imu_lin_acc: torch.Tensor,
    imu_ang_vel: torch.Tensor,
    ## Robot descriptors
    forward_drive_indices: List[int],
) -> StepReturn:
    num_envs = episode_length.size(0)
    # dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Transforms (world frame)
    # Robot -> Target
    tf_pos_robot_to_target = tf_pos_robot[:, :2] - tf_pos_target[:, :2]
    dist_robot_to_target = torch.norm(tf_pos_robot_to_target, dim=-1)

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Angular velocity
    WEIGHT_ANGULAR_VELOCITY = -0.25
    MAX_ANGULAR_VELOCITY_PENALTY = -5.0
    penalty_angular_velocity = torch.clamp_min(
        WEIGHT_ANGULAR_VELOCITY * torch.sum(torch.square(vel_ang_robot[:, :2]), dim=1),
        min=MAX_ANGULAR_VELOCITY_PENALTY,
    )

    # Penalty: Distance | Robot <--> Target
    WEIGHT_DISTANCE_ROBOT_TO_TARGET = -16.0
    penalty_distance_robot_to_target = (
        WEIGHT_DISTANCE_ROBOT_TO_TARGET * dist_robot_to_target
    )

    # Penalty: Backward motion
    WEIGHT_BACKWARD_MOTION = -0.5
    penalty_backward_motion = WEIGHT_BACKWARD_MOTION * torch.norm(
        torch.clamp_max(act_current[:, forward_drive_indices], 0.0), dim=-1
    )

    # Reward: Distance | Robot <--> Target (precision)
    WEIGHT_DISTANCE_ROBOT_TO_TARGET_PRECISION = 64.0
    TANH_STD_DISTANCE_ROBOT_TO_TARGET_PRECISION = 0.05
    reward_distance_robot_to_target_precision = (
        WEIGHT_DISTANCE_ROBOT_TO_TARGET_PRECISION
        * (
            1.0
            - torch.tanh(
                dist_robot_to_target / TANH_STD_DISTANCE_ROBOT_TO_TARGET_PRECISION
            )
        )
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "vel_lin_robot": vel_lin_robot,
                "vel_ang_robot": vel_ang_robot,
                "tf_pos_robot_to_target": tf_pos_robot_to_target,
            },
            "proprio": {
                "imu_lin_acc": imu_lin_acc,
                "imu_ang_vel": imu_ang_vel,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_angular_velocity": penalty_angular_velocity,
            "penalty_distance_robot_to_target": penalty_distance_robot_to_target,
            "penalty_backward_motion": penalty_backward_motion,
            "reward_distance_robot_to_target_precision": reward_distance_robot_to_target_precision,
        },
        termination,
        truncation,
    )



from srb.assets.object.tool import Kinova300
from srb.core.action import (  # noqa: F401
    ActionGroup,
    DifferentialIKControllerCfg,
    DifferentialInverseKinematicsActionCfg,
    InverseKinematicsActionGroup,
    OperationalSpaceControlActionGroup,
    OperationalSpaceControllerActionCfg,
    OperationalSpaceControllerCfg,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, SerialManipulator, Tool, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.nucleus import ISAAC_NUCLEUS_DIR
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class KinovaJ2n6s(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/j2n6s",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("manipulator")
            .joinpath("kinova_j2n6s.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "j2n6s_joint_1": 0.0,
                "j2n6s_joint_2": deg_to_rad(150),
                "j2n6s_joint_3": deg_to_rad(330),
                "j2n6s_joint_4": deg_to_rad(-90),
                "j2n6s_joint_5": 0.0,
                "j2n6s_joint_6": 0.0,
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=["j2n6s_joint_[1-6]"],
                velocity_limit=100.0,
                effort_limit={
                    "j2n6s_joint_[1-2]": 80.0,
                    "j2n6s_joint_3": 40.0,
                    "j2n6s_joint_[4-6]": 20.0,
                },
                stiffness={
                    "j2n6s_joint_[1-3]": 4000.0,
                    "j2n6s_joint_[4-6]": 1500.0,
                },
                damping={
                    "j2n6s_joint_[1-3]": 1000.0,
                    "j2n6s_joint_[4-6]": 500.0,
                },
            ),
        },
    )
    end_effector: Tool | None = Kinova300()

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=["j2n6s_joint_[1-6]"],
            base_name="j2n6s_link_base",
            body_name="j2n6s_link_6",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="j2n6s_link_base")
    frame_flange: Frame = Frame(
        prim_relpath="j2n6s_link_6",
        offset=Transform(
            rot=rpy_to_quat(180.0, 0.0, 0.0),
        ),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="j2n6s_link_base/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="j2n6s_link_6/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


class KinovaJ2n7s(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/j2n7s",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("manipulator")
            .joinpath("kinova_j2n7s.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "j2n7s_joint_1": 0.0,
                "j2n7s_joint_2": deg_to_rad(150),
                "j2n7s_joint_3": 0.0,
                "j2n7s_joint_4": deg_to_rad(30),
                "j2n7s_joint_5": deg_to_rad(90),
                "j2n7s_joint_6": 0.0,
                "j2n7s_joint_7": 0.0,
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=["j2n7s_joint_[1-7]"],
                velocity_limit=100.0,
                effort_limit={
                    "j2n7s_joint_[1-2]": 8000.0,
                    "j2n7s_joint_[3-4]": 4000.0,
                    "j2n7s_joint_[5-7]": 2000.0,
                },
                stiffness={
                    "j2n7s_joint_[1-4]": 4000.0,
                    "j2n7s_joint_[5-7]": 1000.0,
                },
                damping={
                    "j2n7s_joint_[1-4]": 1000.0,
                    "j2n7s_joint_[5-7]": 500.0,
                },
            ),
        },
    )
    end_effector: Tool | None = Kinova300()

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=["j2n7s_joint_[1-7]"],
            base_name="j2n7s_link_base",
            body_name="j2n7s_link_7",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="j2n7s_link_base")
    frame_flange: Frame = Frame(
        prim_relpath="j2n7s_link_7",
        offset=Transform(
            rot=rpy_to_quat(180.0, 0.0, 0.0),
        ),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="j2n7s_link_base/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="j2n7s_link_7/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )


# TODO[low]: Fix Kinova Gen3 with end-effectors (no idea why it fails to find rigid body/articulation for end-effector)
class KinovaGen3n7(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/kinova_gen3",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/Kinova/Gen3/gen3n7_instanceable.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "joint_1": 0.0,
                "joint_2": deg_to_rad(45),
                "joint_3": 0.0,
                "joint_4": deg_to_rad(45),
                "joint_5": 0.0,
                "joint_6": deg_to_rad(90),
                "joint_7": 0.0,
            },
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=["joint_[1-7]"],
                effort_limit={
                    "joint_[1-4]": 2340.0,
                    "joint_[5-7]": 540.0,
                },
                velocity_limit=10.0,
                stiffness=1000.0,
                damping=100.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=["joint_[1-7]"],
            base_name="base_link",
            body_name="end_effector_link",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base_link")
    frame_flange: Frame = Frame(
        prim_relpath="end_effector_link",
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="base_link/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="end_effector_link/camera_wrist",
        offset=Transform(
            pos=(0.07, 0.0, 0.05),
            rot=rpy_to_quat(0.0, -60.0, 180.0),
        ),
    )



from .combined import (  # noqa: F401
    GenericAerialManipulator,
    GenericGroundManipulator,
    GenericOrbitalManipulator,
)
from .generic_humanoid import Humanoid21, Humanoid28  # noqa: F401
from .unitree import UnitreeG1, UnitreeH1  # noqa: F401



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset import ArticulationCfg
from srb.core.asset.common import Frame
from srb.core.asset.robot.mobile.mobile_robot import MobileRobot, MobileRobotRegistry
from srb.core.asset.robot.mobile.mobile_robot_type import MobileRobotType


class GroundRobot(MobileRobot, mobile_robot_entrypoint=MobileRobotType.GROUND):
    ## Model
    asset_cfg: ArticulationCfg

    ## Frames
    frame_front_camera: Frame

    @classmethod
    def mobile_robot_registry(cls) -> Sequence[Type[GroundRobot]]:
        return MobileRobotRegistry.registry.get(MobileRobotType.GROUND, [])  # type: ignore


class WheeledRobot(GroundRobot, mobile_robot_metaclass=True):
    pass


class LeggedRobot(GroundRobot, mobile_robot_metaclass=True):
    ## Links
    regex_feet_links: str



from dataclasses import MISSING
from typing import TYPE_CHECKING, Dict, Literal, Sequence, Tuple

from srb.core.sensor import CameraCfg
from srb.utils.cfg import configclass

if TYPE_CHECKING:
    from srb._typing import AnyEnvCfg


@configclass
class VisualExtCfg:
    ## Extension overrides
    # Never stack visual environments
    stack: bool = False
    # Use a skydome for the environment
    skydome: Literal["low_res", "high_res"] | bool | None = True
    # Re-render frames on reset for visual observations
    rerender_on_reset: bool = True
    # Always disable debug visualization
    debug_vis: bool = False

    ## Camera sensors
    cameras_cfg: Dict[str, CameraCfg] = MISSING  # type: ignore
    camera_resolution: Tuple[int, int] | None = (64, 64)
    camera_framerate: float = 0.0
    camera_data_types: (
        Sequence[
            Literal[
                "rgb",  # same as "rgba"
                "depth",  # same as "distance_to_image_plane"
                "distance_to_camera",
                "normals",
                "motion_vectors",
                "semantic_segmentation",
                "instance_segmentation_fast",
                "instance_id_segmentation_fast",
                # "instance_segmentation",
                # "instance_id_segmentation",
                # "bounding_box_2d_tight",
                # "bounding_box_2d_tight_fast",
                # "bounding_box_2d_loose",
                # "bounding_box_2d_loose_fast",
                # "bounding_box_3d",
                # "bounding_box_3d_fast",
            ]
        ]
        | None
    ) = ("rgb", "depth")

    def wrap(self, env_cfg: "AnyEnvCfg"):
        ## Add camera sensors to the scene
        for camera_key, camera_cfg in self.cameras_cfg.items():
            if self.camera_resolution is not None:
                camera_cfg.width = self.camera_resolution[0]
                camera_cfg.height = self.camera_resolution[1]
            if self.camera_framerate is not None:
                camera_cfg.update_period = self.camera_framerate
            if self.camera_data_types is not None:
                camera_cfg.data_types = self.camera_data_types  # type: ignore
            setattr(env_cfg.scene, camera_key, camera_cfg)



from typing import Any, List, Literal, Mapping, Type

import gymnasium

from srb.utils.cfg import parse_algo_configs
from srb.utils.path import SRB_HYPERPARAMS_DIR


def register_srb_tasks(
    tasks: Mapping[
        str,
        Mapping[
            Literal["entry_point", "task_cfg", "cfg_dir"],
            gymnasium.Env | Any | str,
        ],
    ],
    *,
    default_entry_point: Type[gymnasium.Env] | None = None,
    default_task_cfg: Any | None = None,
    default_cfg_dir: str | None = SRB_HYPERPARAMS_DIR.as_posix(),
):
    for id, cfg in tasks.items():
        entry_point: gymnasium.Env = cfg.get("entry_point", default_entry_point)  # type: ignore
        gymnasium.register(
            id=f"srb/{id}",
            entry_point=f"{entry_point.__module__}:{entry_point.__name__}",  # type: ignore
            kwargs={
                "task_cfg": cfg.get("task_cfg", default_task_cfg),
                **parse_algo_configs(cfg.get("cfg_dir", default_cfg_dir)),  # type: ignore
            },
            disable_env_checker=True,
        )


def get_srb_tasks() -> List[str]:
    return [env_id for env_id in gymnasium.registry.keys() if env_id.startswith("srb/")]



from functools import cached_property
from typing import Any, Mapping, Sequence, Tuple

import gymnasium
import torch
from skrl.envs.wrappers.torch import IsaacLabWrapper
from skrl.utils.spaces.torch import (
    flatten_tensorized_space,
    tensorize_space,
    unflatten_tensorized_space,
)


class SkrlEnvWrapper(IsaacLabWrapper):
    def __init__(
        self,
        env: Any,
        obs_keys: Sequence[str] = [],
        state_keys: Sequence[str] | None = None,
    ) -> None:
        super().__init__(env)
        self._obs_keys = obs_keys
        self._state_keys = state_keys

        self._clip_actions_min = torch.tensor(
            self.action_space.low,  # type: ignore
            device=self.device,
            dtype=torch.float32,
        )
        self._clip_actions_max = torch.tensor(
            self.action_space.high,  # type: ignore
            device=self.device,
            dtype=torch.float32,
        )

    @cached_property
    def action_space(self) -> gymnasium.Space:
        return gymnasium.spaces.Box(
            low=-1.0, high=1.0, shape=super().action_space.shape
        )

    @cached_property
    def observation_space(self) -> gymnasium.Space:
        if hasattr(self._unwrapped, "single_observation_space"):
            obs_space = self._unwrapped.single_observation_space
        else:
            obs_space = self._unwrapped.observation_space

        if self._obs_keys:
            return gymnasium.spaces.Dict(
                {key: obs_space[key] for key in self._obs_keys}
            )
        else:
            return obs_space

    @cached_property
    def state_space(self) -> gymnasium.Space | None:
        """State space"""
        if hasattr(self._unwrapped, "state_space"):
            return self._unwrapped.state_space

        if hasattr(self._unwrapped, "single_observation_space"):
            obs_space = self._unwrapped.single_observation_space
        else:
            obs_space = self._unwrapped.observation_space

        if self._state_keys is None:
            return None
        elif self._state_keys:
            return gymnasium.spaces.Dict(
                {key: obs_space[key] for key in self._state_keys}
            )
        else:
            return obs_space

    def step(
        self, actions: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, Any]:
        actions = torch.clamp(
            actions, min=self._clip_actions_min, max=self._clip_actions_max
        )
        actions = unflatten_tensorized_space(self.action_space, actions)
        observations, reward, terminated, truncated, self._info = self._env.step(
            actions
        )
        self._observations = flatten_tensorized_space(
            tensorize_space(
                self.observation_space, self.__extract_observations(observations)
            )
        )
        return (
            self._observations,
            reward.view(-1, 1),
            terminated.view(-1, 1),
            truncated.view(-1, 1),
            self._info,
        )

    def reset(self) -> Tuple[torch.Tensor, Any]:
        if self._reset_once:
            observations, self._info = self._env.reset()
            self._observations = flatten_tensorized_space(
                tensorize_space(
                    self.observation_space, self.__extract_observations(observations)
                )
            )
            self._reset_once = False
        return self._observations, self._info

    def __extract_observations(
        self, observations: Mapping[str, torch.Tensor]
    ) -> Mapping[str, torch.Tensor] | torch.Tensor:
        if not self._obs_keys:
            return observations
        return {key: observations[key] for key in self._obs_keys}
